from dataclasses import dataclass, field
from .constant import RANGE_INVALID, RANGE_OVERWORLD
from ..internal.symbol_export_chunk import (
    chunk_biome,
    chunk_block,
    chunk_blocks,
    chunk_compact,
    chunk_equals,
    chunk_highest_filled_sub_chunk,
    chunk_range,
    chunk_set_biome,
    chunk_set_block,
    chunk_set_blocks,
    chunk_sub,
    chunk_sub_chunk,
    chunk_sub_index,
    chunk_sub_y,
    new_chunk as nc,
    release_chunk,
)
from ..world.define import QuickChunkBlocks, Range
from ..world.sub_chunk import SubChunk


class ChunkBase:
    """ChunkBase is the base implement of a Minecraft chunk."""

    _chunk_id: int

    def __init__(self):
        self._chunk_id = -1

    def __del__(self):
        if self._chunk_id >= 0 and not release_chunk is None:
            release_chunk(self._chunk_id)

    def is_valid(self) -> bool:
        """
        is_valid check current chunk is valid or not.
        If not valid, it means the chunk is actually not exist, not only Python but also in Go.
        Try to use a invalid chunk is not allowed, and any operation will be terminated.

        Returns:
            bool: Wheatear the chunk is valid or not.
        """
        return self._chunk_id >= 0


@dataclass
class Chunk(ChunkBase):
    """
    Chunk is a segment in the world with a size of 16x16x256 blocks. A chunk contains multiple sub chunks
    and stores other information such as biomes.
    It is not safe to call methods on Chunk simultaneously from multiple goroutines.
    """

    _range: Range = field(default_factory=lambda: Range())
    _known_range: bool = False

    def __init__(self):
        super().__init__()

    def biome(self, x: int, y: int, z: int) -> int:
        """biome returns the biome ID at a specific column in the chunk.

        Args:
            x (int): The relative x position of this column. Must in a range of 0-15.
            y (int): The y position of this column.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this column. Must in a range of 0-15.

        Returns:
            int: The biome ID of this column.
                 If current chunk is not found, then return -1.
        """
        return chunk_biome(self._chunk_id, x, y, z)

    def block(self, x: int, y: int, z: int, layer: int) -> int:
        """Block returns the runtime ID of the block at a given x, y and z in a chunk at the given layer.

        Args:
            x (int): The relative x position of this block. Must in a range of 0-15.
            y (int): The y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this block. Must in a range of 0-15.
            layer (int): The layer to find this block.

        Returns:
            int: Return the block runtime ID of target block.
                 If current chunk is not found, then return -1.
                 Note that if no sub chunk exists at the given y, the block is assumed to be air.
        """
        return chunk_block(self._chunk_id, x, y, z, layer)

    def blocks(self, layer: int) -> QuickChunkBlocks:
        """
        blocks returns all blocks (block runtime ids) whose in the target layer of this chunk.
        It is highly suggest you use this instead of c.block(...) if you are trying to query so
        many blocks from this chunk.

        Args:
            layer (int): The layer of the blocks in this chunk that you want to find.

        Returns:
            QuickChunkBlocks: All blocks of the target layer in this chunk if current chunk is exist.
                              If the target layer is not exist, then you get a chunk full of air.
                              Note that this implement don't do further check (failed to get range of
                              current chunk or underlying blocks list is empty) due to this is aims to
                              increase block query/set speed, and you should take responsibility for
                              any possible error.
        """
        if not self._known_range:
            self._range = self.range()
        return QuickChunkBlocks(
            chunk_blocks(self._chunk_id, layer),
            self._range.start_range,
            self._range.end_range,
        )

    def compact(self):
        """
        compact compacts the chunk as much as possible, getting rid of any sub chunks that are empty,
        and compacts all storages in the sub chunks to occupy as little space as possible.
        compact should be called right before the chunk is saved in order to optimise the storage space.

        Raises:
            Exception: When failed to compact.
        """
        err = chunk_compact(self._chunk_id)
        if len(err) > 0:
            raise Exception(err)

    def equals(self, another_chunk: ChunkBase) -> bool:
        """equals returns if the chunk passed is equal to the current one.

        Args:
            another_chunk (ChunkBase): The chunk passed.

        Returns:
            bool: The compare result.
                  True for the contents of two chunk is the same.
                  False for current chunk or another_chunk is not found.
        """
        result = chunk_equals(self._chunk_id, another_chunk._chunk_id)
        return result == 1

    def highest_filled_sub_chunk(self) -> int:
        """
        highest_filled_sub_chunk returns the index of the highest sub chunk in the chunk
        that has any blocks in it. 0 is returned if no subchunks have any blocks.

        Returns:
            int: The index of the highest sub chunk in the chunk that has any blocks in it.
                 If no subchunks have any block, then return 0.
                 Additionally, if current chunk is not found, then return -1.
        """
        return chunk_highest_filled_sub_chunk(self._chunk_id)

    def range(self) -> Range:
        """Range returns the Range of the Chunk as passed to new_chunk.

        Returns:
            Range: If current chunk is not found, return RANGE_INVALID.
                   Otherwise, return the range and True.
        """
        if self._known_range:
            return self._range

        start_range, end_range, ok = chunk_range(self._chunk_id)
        if not ok:
            return RANGE_INVALID

        self._known_range = True
        self._range = Range(start_range, end_range)

        return self._range

    def set_biome(self, x: int, y: int, z: int, biome_id: int):
        """set_biome sets the biome ID at a specific column in the chunk.

        Args:
            x (int): The relative x position of this column. Must in a range of 0-15.
            y (int): The y position of this column.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this column. Must in a range of 0-15.
            biome_id (int): The biome ID want to set.

        Raises:
            Exception: When failed to set biome ID.
        """
        err = chunk_set_biome(self._chunk_id, x, y, z, biome_id)
        if len(err) > 0:
            raise Exception(err)

    def set_block(self, x: int, y: int, z: int, layer: int, block_runtime_id: int):
        """
        set_block sets the runtime ID of a block at a given x, y and z in a chunk at the given layer.
        If no sub chunk exists at the given y, a new sub chunk is created and the block is set.

        Args:
            x (int): The relative x position of this block. Must in a range of 0-15.
            y (int): The y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this block. Must in a range of 0-15.
            layer (int): The layer that this block in.
            block_runtime_id (int): The result block that this block will be.

        Raises:
            Exception: When failed to set block.
        """
        err = chunk_set_block(self._chunk_id, x, y, z, layer, block_runtime_id)
        if len(err) > 0:
            raise Exception(err)

    def set_blocks(self, layer: int, blocks: QuickChunkBlocks):
        """
        set_blocks sets the whole chunk blocks in layer by given block runtime ids.

        It is highly suggest you use this instead of c.set_block(...) if you are trying to modify
        so many blocks to this chunk.

        Note that this implement will not check the underlying blocks list is valid
        or not due to this is aims to increase block query/set speed, and you should
        take responsibility for any possible error.

        Args:
            layer (int): The blocks in the target layer of this chunk that you want to overwrite.

        Raises:
            Exception: When failed to set blocks.
        """
        err = chunk_set_blocks(self._chunk_id, layer, blocks.blocks)
        if len(err) > 0:
            raise Exception(err)

    def sub(self) -> list[SubChunk]:
        """
        sub returns a list of all sub chunks present in the chunk.
        Note that after edit those sub chunk, you just only need to save this chunk,
        but not need to save the modified sub chunks.

        Returns:
            list[SubChunk]: All sub chunks present in the chunk.
                            If current chunk is not found, or this chunk have no sub chunk, then return empty list.
        """
        result = []
        for i in chunk_sub(self._chunk_id):
            s = SubChunk()
            s._sub_chunk_id = i
            result.append(s)
        return result

    def sub_chunk(self, y: int) -> SubChunk:
        """
        sub_chunk finds the correct sub chunk in the Chunk by a y position.
        Note that after edit this sub chunk, you just only need to save this chunk,
        but not need to save the modified sub chunks.

        Args:
            y (int): The y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).

        Returns:
            SubChunk: If current chunk is not found, then return a invalid sub chunk.
                      Otherwise, return the target sub chunk and True.
                      Note that you could use s.is_valid() to check whether the sub chunk is valid or not.
        """
        s = SubChunk()
        s._sub_chunk_id = chunk_sub_chunk(self._chunk_id, y)
        return s

    def sub_index(self, y: int) -> int:
        """sub_index returns the sub chunk index matching the y position passed.

        Args:
            y (int): The relative y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).

        Returns:
            int: The index of the y position.
                 If current sub chunk is not found, then return -1.
        """
        return chunk_sub_index(self._chunk_id, y)

    def sub_y(self, index: int) -> int:
        """sub_y returns the sub chunk y value matching the index passed.

        Args:
            index (int): The given index that used to compute the value of y.

        Returns:
            int: The y position who could match the given index.
                 y is in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
                 If current sub chunk is not found, then return -1.
        """
        return chunk_sub_y(self._chunk_id, index)


def new_chunk(r: Range = RANGE_OVERWORLD) -> Chunk:
    """NewChunk initialises a new chunk who full of air and returns it, so that it may be used.

    Args:
        r (Range, optional): The Y range of this chunk could reach. Defaults to RANGE_OVERWORLD.

    Returns:
        Chunk: A new chunk.
    """
    c = Chunk()
    c._chunk_id = nc(r.start_range, r.end_range)
    return c



import struct, nbtlib
from io import BytesIO
from .types import LIB
from .types import CSlice, CString, CInt
from .types import as_c_bytes, as_python_bytes, as_c_string
from ..utils import marshalNBT, unmarshalNBT


LIB.RuntimeIDToState.argtypes = [CInt]
LIB.StateToRuntimeID.argtypes = [CString, CSlice]
LIB.SubChunkNetworkPayload.argtypes = [CInt, CInt, CInt, CInt]
LIB.FromSubChunkNetworkPayload.argtypes = [CInt, CInt, CSlice]
LIB.SubChunkDiskPayload.argtypes = [CInt, CInt, CInt, CInt]
LIB.FromSubChunkDiskPayload.argtypes = [CInt, CInt, CSlice]

LIB.RuntimeIDToState.restype = CSlice
LIB.StateToRuntimeID.restype = CSlice
LIB.SubChunkNetworkPayload.restype = CSlice
LIB.FromSubChunkNetworkPayload.restype = CSlice
LIB.SubChunkDiskPayload.restype = CSlice
LIB.FromSubChunkDiskPayload.restype = CSlice


def runtime_id_to_state(
    block_runtime_id: int,
) -> tuple[str, nbtlib.tag.Compound | None, bool]:
    payload = as_python_bytes(LIB.RuntimeIDToState(CInt(block_runtime_id)))
    reader = BytesIO(payload)

    if reader.read(1) == b"\x00":
        return ("", None, False)

    l: int = struct.unpack("<H", reader.read(2))[0]
    name = reader.read(l).decode(encoding="utf-8")

    l = struct.unpack("<I", reader.read(4))[0]
    states_nbt = reader.read(l)

    return (
        name,
        unmarshalNBT.UnMarshalBufferToPythonNBTObject(BytesIO(states_nbt))[0],  # type: ignore
        True,
    )


def state_to_runtime_id(
    block_name: str, block_states: nbtlib.tag.Compound
) -> tuple[int, bool]:
    writer = BytesIO()
    marshalNBT.MarshalPythonNBTObjectToWriter(writer, block_states, "")

    payload = as_python_bytes(
        LIB.StateToRuntimeID(as_c_string(block_name), as_c_bytes(writer.getvalue()))
    )
    reader = BytesIO(payload)

    if reader.read(1) == b"\x00":
        return (0, False)

    return (struct.unpack("<I", reader.read(4))[0], True)


def sub_chunk_network_payload(
    id: int, range_start: int, range_end: int, ind: int
) -> bytes:
    return as_python_bytes(
        LIB.SubChunkNetworkPayload(
            CInt(id), CInt(range_start), CInt(range_end), CInt(ind)
        )
    )


def from_sub_chunk_network_payload(
    range_start: int, range_end: int, payload: bytes
) -> tuple[int, int, bool]:
    reader = BytesIO(
        as_python_bytes(
            LIB.FromSubChunkNetworkPayload(
                CInt(range_start), CInt(range_end), as_c_bytes(payload)
            )
        )
    )

    if reader.read(1) == b"\x00":
        return (0, 0, False)

    index = reader.read(1)[0]
    sub_chunk_id = struct.unpack("<I", reader.read(4))[0]

    return (index, sub_chunk_id, True)


def sub_chunk_disk_payload(
    id: int, range_start: int, range_end: int, ind: int
) -> bytes:
    return as_python_bytes(
        LIB.SubChunkDiskPayload(CInt(id), CInt(range_start), CInt(range_end), CInt(ind))
    )


def from_sub_chunk_disk_payload(
    range_start: int, range_end: int, payload: bytes
) -> tuple[int, int, bool]:
    reader = BytesIO(
        as_python_bytes(
            LIB.FromSubChunkDiskPayload(
                CInt(range_start), CInt(range_end), as_c_bytes(payload)
            )
        )
    )

    if reader.read(1) == b"\x00":
        return (0, 0, False)

    index = reader.read(1)[0]
    sub_chunk_id = struct.unpack("<I", reader.read(4))[0]

    return (index, sub_chunk_id, True)



import nbtlib
import time as Time
from dataclasses import dataclass, field


@dataclass
class Abilities:
    attack_mobs: bool = True
    attack_players: bool = True
    build: bool = True
    mine: bool = True
    doors_and_switches: bool = True
    fly_speed: float = 0.05
    flying: bool = False
    instant_build: bool = False
    invulnerable: bool = False
    lightning: bool = False
    mayFly: bool = False
    op: bool = False
    open_containers: bool = True
    permissions_level: int = 0
    player_permissions_level: int = 1
    teleport: bool = False
    walk_speed: float = 0.1
    verticalFly_speed: float = 1.0

    def marshal(self) -> nbtlib.tag.Compound:
        """marshal encode abilities to its NBT represent

        Returns:
            Compound: The NBT represent of abilities
        """
        return nbtlib.tag.Compound(
            {
                "attackmobs": nbtlib.tag.Byte(self.attack_mobs),
                "attackplayers": nbtlib.tag.Byte(self.attack_players),
                "build": nbtlib.tag.Byte(self.build),
                "mine": nbtlib.tag.Byte(self.mine),
                "doorsandswitches": nbtlib.tag.Byte(self.doors_and_switches),
                "flySpeed": nbtlib.tag.Float(self.fly_speed),
                "flying": nbtlib.tag.Byte(self.flying),
                "instabuild": nbtlib.tag.Byte(self.instant_build),
                "invulnerable": nbtlib.tag.Byte(self.invulnerable),
                "lightning": nbtlib.tag.Byte(self.lightning),
                "mayfly": nbtlib.tag.Byte(self.mayFly),
                "op": nbtlib.tag.Byte(self.op),
                "opencontainers": nbtlib.tag.Byte(self.open_containers),
                "permissionsLevel": nbtlib.tag.Int(self.permissions_level),
                "playerPermissionsLevel": nbtlib.tag.Int(self.player_permissions_level),
                "teleport": nbtlib.tag.Byte(self.teleport),
                "walkSpeed": nbtlib.tag.Float(self.walk_speed),
                "verticalFlySpeed": nbtlib.tag.Float(self.verticalFly_speed),
            }
        )

    def unmarshal(self, compound: nbtlib.tag.Compound):
        """unmarshal read data from compound to initialize the field in abilities.

        Args:
            compound (nbtlib.tag.Compound): The given compound.
        """
        self.attack_mobs = bool(compound["attackmobs"])
        self.attack_players = bool(compound["attackplayers"])
        self.build = bool(compound["build"])
        self.mine = bool(compound["mine"])
        self.doors_and_switches = bool(compound["doorsandswitches"])
        self.fly_speed = float(compound["flySpeed"])
        self.flying = bool(compound["flying"])
        self.instant_build = bool(compound["instabuild"])
        self.invulnerable = bool(compound["invulnerable"])
        self.lightning = bool(compound["lightning"])
        self.mayFly = bool(compound["mayfly"])
        self.op = bool(compound["op"])
        self.open_containers = bool(compound["opencontainers"])
        self.permissions_level = int(compound["permissionsLevel"])
        self.player_permissions_level = int(compound["playerPermissionsLevel"])
        self.teleport = bool(compound["teleport"])
        self.walk_speed = float(compound["walkSpeed"])
        self.verticalFly_speed = float(compound["verticalFlySpeed"])


@dataclass
class LevelDat:
    """
    LevelDat holds a collection of data that specify a range of Settings of the
    world. These Settings usually alter the way that players interact with the
    world. The data held here is usually saved in a level.dat file of the world.
    LevelDat may be used in LevelDat.Unmarshal to collect the data of the level.dat.
    """

    base_game_version: str = "*"
    biome_override: str = ""
    confirmed_platform_locked_content: bool = False
    center_maps_to_origin: bool = False
    cheats_enabled: bool = False
    daylight_cycle: int = 0
    difficulty: int = 2
    edu_offer: int = 0
    flat_world_layers: str = ""
    force_game_type: bool = False
    game_type: int = 1
    generator: int = 2
    inventory_version: str = "1.20.50"
    LANB_roadcast: bool = True
    LANB_roadcast_intent: bool = True
    last_played: int = 0
    level_name: str = "World"
    limited_world_origin_x: int = 0
    limited_world_origin_y: int = 32767
    limited_world_origin_z: int = 0
    limited_world_depth: int = 16
    limited_world_width: int = 16
    minimum_compatible_client_version: tuple[int, int, int, int, int] = (
        1,
        20,
        50,
        0,
        0,
    )
    multi_player_game: bool = True
    multi_player_game_intent: bool = True
    nether_scale: int = 8
    network_version: int = 630
    platform: int = 2
    platform_broadcast_intent: int = 3
    random_seed: int = field(default_factory=lambda: int(Time.time()))
    show_tags: bool = True
    single_use_world: bool = False
    spawn_x: int = 0
    spawn_y: int = 32767
    spawn_z: int = 0
    spawn_v1_villagers: bool = False
    storage_version: int = 9
    time: int = 0
    XBL_broadcast: bool = False
    XBL_broadcast_intent: int = 3
    XBL_broadcast_mode: int = 0
    abilities: Abilities = field(default_factory=lambda: Abilities())
    bonus_chest_enabled: bool = False
    bonus_chest_spawned: bool = False
    command_block_output: bool = True
    command_blocks_enabled: bool = True
    commands_enabled: bool = True
    current_tick: int = 0
    do_day_light_cycle: bool = True
    do_entity_drops: bool = True
    do_fire_tick: bool = True
    do_immediate_respawn: bool = False
    do_insomnia: bool = True
    do_mob_loot: bool = True
    do_mob_spawning: bool = True
    do_tile_drops: bool = True
    do_weather_cycle: bool = True
    drowning_damage: bool = True
    edu_level: bool = False
    education_features_enabled: bool = False
    experimental_game_play: bool = False
    fall_damage: bool = True
    fire_damage: bool = True
    function_command_limit: int = 10000
    has_been_loaded_in_creative: bool = True
    has_locked_behaviour_pack: bool = False
    has_locked_resource_pack: bool = False
    immutable_world: bool = False
    is_created_in_editor: bool = False
    is_exported_from_editor: bool = False
    is_from_locked_template: bool = False
    is_from_world_template: bool = False
    is_world_template_option_locked: bool = False
    keep_inventory: bool = False
    last_opened_with_version: tuple[int, int, int, int, int] = (1, 20, 50, 0, 0)
    lightning_level: float = 1.0
    lightning_time: int = 0
    max_command_chain_length: int = 65535
    mob_griefing: bool = True
    natural_regeneration: bool = True
    prid: str = ""
    pvp: bool = True
    rain_level: float = 1.0
    rain_time: int = 0
    random_tick_speed: int = 1
    requires_copied_pack_removal_check: bool = False
    send_command_feedback: bool = True
    server_chunk_tick_range: int = 6
    show_coordinates: bool = False
    show_death_messages: bool = True
    spawn_mobs: bool = True
    spawn_radius: int = 5
    start_with_map_enabled: bool = False
    texture_packs_required: bool = False
    tnt_explodes: bool = True
    use_msa_gamer_tags_only: bool = False
    world_start_count: int = 0
    experiments: nbtlib.tag.Compound = field(
        default_factory=lambda: nbtlib.tag.Compound()
    )
    freeze_damage: bool = True
    world_policies: nbtlib.tag.Compound = field(
        default_factory=lambda: nbtlib.tag.Compound()
    )
    world_version: int = 1
    respawn_blocks_explode: bool = True
    show_border_effect: bool = True
    permissions_level: int = 0
    player_permissions_level: int = 0
    is_random_seed_allowed: bool = False
    do_limited_crafting: bool = False
    editor_world_type: int = 0
    players_sleeping_percentage: int = 0
    recipes_unlock: bool = False
    natural_generation: bool = False
    projectiles_can_break_blocks: bool = False
    show_recipe_messages: bool = False
    is_hardcore: bool = False
    show_days_played: bool = False
    tnt_explosion_drop_decay: bool = False
    has_uncomplete_world_file_on_disk: bool = False
    player_has_died: bool = False

    def marshal(self) -> nbtlib.tag.Compound:
        """marshal encode level dat to its NBT represent

        Returns:
            Compound: The NBT represent of level dat
        """
        return nbtlib.tag.Compound(
            {
                "baseGameVersion": nbtlib.tag.String(self.base_game_version),
                "BiomeOverride": nbtlib.tag.String(self.biome_override),
                "ConfirmedPlatformLockedContent": nbtlib.tag.Byte(
                    self.confirmed_platform_locked_content
                ),
                "CenterMapsToOrigin": nbtlib.tag.Byte(self.center_maps_to_origin),
                "cheatsEnabled": nbtlib.tag.Byte(self.cheats_enabled),
                "daylightCycle": nbtlib.tag.Int(self.daylight_cycle),
                "Difficulty": nbtlib.tag.Int(self.difficulty),
                "eduOffer": nbtlib.tag.Int(self.edu_offer),
                "FlatWorldLayers": nbtlib.tag.String(self.flat_world_layers),
                "ForceGameType": nbtlib.tag.Byte(self.force_game_type),
                "GameType": nbtlib.tag.Int(self.game_type),
                "Generator": nbtlib.tag.Int(self.generator),
                "InventoryVersion": nbtlib.tag.String(self.inventory_version),
                "LANBroadcast": nbtlib.tag.Byte(self.LANB_roadcast),
                "LANBroadcastIntent": nbtlib.tag.Byte(self.LANB_roadcast_intent),
                "LastPlayed": nbtlib.tag.Long(self.last_played),
                "LevelName": nbtlib.tag.String(self.level_name),
                "LimitedWorldOriginX": nbtlib.tag.Int(self.limited_world_origin_x),
                "LimitedWorldOriginY": nbtlib.tag.Int(self.limited_world_origin_y),
                "LimitedWorldOriginZ": nbtlib.tag.Int(self.limited_world_origin_z),
                "limitedWorldDepth": nbtlib.tag.Int(self.limited_world_depth),
                "limitedWorldWidth": nbtlib.tag.Int(self.limited_world_width),
                "MinimumCompatibleClientVersion": nbtlib.tag.List(
                    [
                        nbtlib.tag.Int(self.minimum_compatible_client_version[0]),
                        nbtlib.tag.Int(self.minimum_compatible_client_version[1]),
                        nbtlib.tag.Int(self.minimum_compatible_client_version[2]),
                        nbtlib.tag.Int(self.minimum_compatible_client_version[3]),
                        nbtlib.tag.Int(self.minimum_compatible_client_version[4]),
                    ]
                ),
                "MultiplayerGame": nbtlib.tag.Byte(self.multi_player_game),
                "MultiplayerGameIntent": nbtlib.tag.Byte(self.multi_player_game_intent),
                "NetherScale": nbtlib.tag.Int(self.nether_scale),
                "NetworkVersion": nbtlib.tag.Int(self.network_version),
                "Platform": nbtlib.tag.Int(self.platform),
                "PlatformBroadcastIntent": nbtlib.tag.Int(
                    self.platform_broadcast_intent
                ),
                "RandomSeed": nbtlib.tag.Long(self.random_seed),
                "showtags": nbtlib.tag.Byte(self.show_tags),
                "isSingleUseWorld": nbtlib.tag.Byte(self.single_use_world),
                "SpawnX": nbtlib.tag.Int(self.spawn_x),
                "SpawnY": nbtlib.tag.Int(self.spawn_y),
                "SpawnZ": nbtlib.tag.Int(self.spawn_z),
                "SpawnV1Villagers": nbtlib.tag.Byte(self.spawn_v1_villagers),
                "StorageVersion": nbtlib.tag.Int(self.storage_version),
                "Time": nbtlib.tag.Long(self.time),
                "XBLBroadcast": nbtlib.tag.Byte(self.XBL_broadcast),
                "XBLBroadcastIntent": nbtlib.tag.Int(self.XBL_broadcast_intent),
                "XBLBroadcastMode": nbtlib.tag.Int(self.XBL_broadcast_mode),
                "abilities": self.abilities.marshal(),
                "bonusChestEnabled": nbtlib.tag.Byte(self.bonus_chest_enabled),
                "bonusChestSpawned": nbtlib.tag.Byte(self.bonus_chest_spawned),
                "commandblockoutput": nbtlib.tag.Byte(self.command_block_output),
                "commandblocksenabled": nbtlib.tag.Byte(self.command_blocks_enabled),
                "commandsEnabled": nbtlib.tag.Byte(self.commands_enabled),
                "currentTick": nbtlib.tag.Long(self.current_tick),
                "dodaylightcycle": nbtlib.tag.Byte(self.do_day_light_cycle),
                "doentitydrops": nbtlib.tag.Byte(self.do_entity_drops),
                "dofiretick": nbtlib.tag.Byte(self.do_fire_tick),
                "doimmediaterespawn": nbtlib.tag.Byte(self.do_immediate_respawn),
                "doinsomnia": nbtlib.tag.Byte(self.do_insomnia),
                "domobloot": nbtlib.tag.Byte(self.do_mob_loot),
                "domobspawning": nbtlib.tag.Byte(self.do_mob_spawning),
                "dotiledrops": nbtlib.tag.Byte(self.do_tile_drops),
                "doweathercycle": nbtlib.tag.Byte(self.do_weather_cycle),
                "drowningdamage": nbtlib.tag.Byte(self.drowning_damage),
                "eduLevel": nbtlib.tag.Byte(self.edu_level),
                "educationFeaturesEnabled": nbtlib.tag.Byte(
                    self.education_features_enabled
                ),
                "experimentalgameplay": nbtlib.tag.Byte(self.experimental_game_play),
                "falldamage": nbtlib.tag.Byte(self.fall_damage),
                "firedamage": nbtlib.tag.Byte(self.fire_damage),
                "functioncommandlimit": nbtlib.tag.Int(self.function_command_limit),
                "hasBeenLoadedInCreative": nbtlib.tag.Byte(
                    self.has_been_loaded_in_creative
                ),
                "hasLockedBehaviorPack": nbtlib.tag.Byte(
                    self.has_locked_behaviour_pack
                ),
                "hasLockedResourcePack": nbtlib.tag.Byte(self.has_locked_resource_pack),
                "immutableWorld": nbtlib.tag.Byte(self.immutable_world),
                "isCreatedInEditor": nbtlib.tag.Byte(self.is_created_in_editor),
                "isExportedFromEditor": nbtlib.tag.Byte(self.is_exported_from_editor),
                "isFromLockedTemplate": nbtlib.tag.Byte(self.is_from_locked_template),
                "isFromWorldTemplate": nbtlib.tag.Byte(self.is_from_world_template),
                "isWorldTemplateOptionLocked": nbtlib.tag.Byte(
                    self.is_world_template_option_locked
                ),
                "keepinventory": nbtlib.tag.Byte(self.keep_inventory),
                "lastOpenedWithVersion": nbtlib.tag.List(
                    [
                        nbtlib.tag.Int(self.last_opened_with_version[0]),
                        nbtlib.tag.Int(self.last_opened_with_version[1]),
                        nbtlib.tag.Int(self.last_opened_with_version[2]),
                        nbtlib.tag.Int(self.last_opened_with_version[3]),
                        nbtlib.tag.Int(self.last_opened_with_version[4]),
                    ]
                ),
                "lightningLevel": nbtlib.tag.Float(self.lightning_level),
                "lightningTime": nbtlib.tag.Int(self.lightning_time),
                "maxcommandchainlength": nbtlib.tag.Int(self.max_command_chain_length),
                "mobgriefing": nbtlib.tag.Byte(self.mob_griefing),
                "naturalregeneration": nbtlib.tag.Byte(self.natural_regeneration),
                "prid": nbtlib.tag.String(self.prid),
                "pvp": nbtlib.tag.Byte(self.pvp),
                "rainLevel": nbtlib.tag.Float(self.rain_level),
                "rainTime": nbtlib.tag.Int(self.rain_time),
                "randomtickspeed": nbtlib.tag.Int(self.random_tick_speed),
                "requiresCopiedPackRemovalCheck": nbtlib.tag.Byte(
                    self.requires_copied_pack_removal_check
                ),
                "sendcommandfeedback": nbtlib.tag.Byte(self.send_command_feedback),
                "serverChunkTickRange": nbtlib.tag.Int(self.server_chunk_tick_range),
                "showcoordinates": nbtlib.tag.Byte(self.show_coordinates),
                "showdeathmessages": nbtlib.tag.Byte(self.show_death_messages),
                "spawnMobs": nbtlib.tag.Byte(self.spawn_mobs),
                "spawnradius": nbtlib.tag.Int(self.spawn_radius),
                "startWithMapEnabled": nbtlib.tag.Byte(self.start_with_map_enabled),
                "texturePacksRequired": nbtlib.tag.Byte(self.texture_packs_required),
                "tntexplodes": nbtlib.tag.Byte(self.tnt_explodes),
                "useMsaGamertagsOnly": nbtlib.tag.Byte(self.use_msa_gamer_tags_only),
                "worldStartCount": nbtlib.tag.Long(self.world_start_count),
                "experiments": self.experiments,
                "freezedamage": nbtlib.tag.Byte(self.freeze_damage),
                "world_policies": self.world_policies,
                "WorldVersion": nbtlib.tag.Int(self.world_version),
                "respawnblocksexplode": nbtlib.tag.Byte(self.respawn_blocks_explode),
                "showbordereffect": nbtlib.tag.Byte(self.show_border_effect),
                "permissionsLevel": nbtlib.tag.Int(self.permissions_level),
                "playerPermissionsLevel": nbtlib.tag.Int(self.player_permissions_level),
                "isRandomSeedAllowed": nbtlib.tag.Byte(self.is_random_seed_allowed),
                "dolimitedcrafting": nbtlib.tag.Byte(self.do_limited_crafting),
                "editorWorldType": nbtlib.tag.Int(self.editor_world_type),
                "playerssleepingpercentage": nbtlib.tag.Int(
                    self.players_sleeping_percentage
                ),
                "recipesunlock": nbtlib.tag.Byte(self.recipes_unlock),
                "naturalgeneration": nbtlib.tag.Byte(self.natural_generation),
                "projectilescanbreakblocks": nbtlib.tag.Byte(
                    self.projectiles_can_break_blocks
                ),
                "showrecipemessages": nbtlib.tag.Byte(self.show_recipe_messages),
                "IsHardcore": nbtlib.tag.Byte(self.is_hardcore),
                "showdaysplayed": nbtlib.tag.Byte(self.show_days_played),
                "tntexplosiondropdecay": nbtlib.tag.Byte(self.tnt_explosion_drop_decay),
                "HasUncompleteWorldFileOnDisk": nbtlib.tag.Byte(
                    self.has_uncomplete_world_file_on_disk
                ),
                "PlayerHasDied": nbtlib.tag.Byte(self.player_has_died),
            }
        )

    def unmarshal(self, compound: nbtlib.tag.Compound):
        """unmarshal read data from compound to initialize the field in level dat.

        Args:
            compound (nbtlib.tag.Compound): The given compound.
        """
        self.base_game_version = str(compound["baseGameVersion"])
        self.biome_override = str(compound["BiomeOverride"])
        self.confirmed_platform_locked_content = bool(
            compound["ConfirmedPlatformLockedContent"]
        )
        self.center_maps_to_origin = bool(compound["CenterMapsToOrigin"])
        self.cheats_enabled = bool(compound["cheatsEnabled"])
        self.daylight_cycle = int(compound["daylightCycle"])
        self.difficulty = int(compound["Difficulty"])
        self.edu_offer = int(compound["eduOffer"])
        self.flat_world_layers = str(compound["FlatWorldLayers"])
        self.force_game_type = bool(compound["ForceGameType"])
        self.game_type = int(compound["GameType"])
        self.generator = int(compound["Generator"])
        self.inventory_version = str(compound["InventoryVersion"])
        self.LANB_roadcast = bool(compound["LANBroadcast"])
        self.LANB_roadcast_intent = bool(compound["LANBroadcastIntent"])
        self.last_played = int(compound["LastPlayed"])
        self.level_name = str(compound["LevelName"])
        self.limited_world_origin_x = int(compound["LimitedWorldOriginX"])
        self.limited_world_origin_y = int(compound["LimitedWorldOriginY"])
        self.limited_world_origin_z = int(compound["LimitedWorldOriginZ"])
        self.limited_world_depth = int(compound["limitedWorldDepth"])
        self.limited_world_width = int(compound["limitedWorldWidth"])
        self.minimum_compatible_client_version = (
            int(compound["MinimumCompatibleClientVersion"][0]),
            int(compound["MinimumCompatibleClientVersion"][1]),
            int(compound["MinimumCompatibleClientVersion"][2]),
            int(compound["MinimumCompatibleClientVersion"][3]),
            int(compound["MinimumCompatibleClientVersion"][4]),
        )
        self.multi_player_game = bool(compound["MultiplayerGame"])
        self.multi_player_game_intent = bool(compound["MultiplayerGameIntent"])
        self.nether_scale = int(compound["NetherScale"])
        self.network_version = int(compound["NetworkVersion"])
        self.platform = int(compound["Platform"])
        self.platform_broadcast_intent = int(compound["PlatformBroadcastIntent"])
        self.random_seed = int(compound["RandomSeed"])
        self.show_tags = bool(compound["showtags"])
        self.single_use_world = bool(compound["isSingleUseWorld"])
        self.spawn_x = int(compound["SpawnX"])
        self.spawn_y = int(compound["SpawnY"])
        self.spawn_z = int(compound["SpawnZ"])
        self.spawn_v1_villagers = bool(compound["SpawnV1Villagers"])
        self.storage_version = int(compound["StorageVersion"])
        self.time = int(compound["Time"])
        self.XBL_broadcast = bool(compound["XBLBroadcast"])
        self.XBL_broadcast_intent = int(compound["XBLBroadcastIntent"])
        self.XBL_broadcast_mode = int(compound["XBLBroadcastMode"])
        self.abilities.unmarshal(compound["abilities"])
        self.bonus_chest_enabled = bool(compound["bonusChestEnabled"])
        self.bonus_chest_spawned = bool(compound["bonusChestSpawned"])
        self.command_block_output = bool(compound["commandblockoutput"])
        self.command_blocks_enabled = bool(compound["commandblocksenabled"])
        self.commands_enabled = bool(compound["commandsEnabled"])
        self.current_tick = int(compound["currentTick"])
        self.do_day_light_cycle = bool(compound["dodaylightcycle"])
        self.do_entity_drops = bool(compound["doentitydrops"])
        self.do_fire_tick = bool(compound["dofiretick"])
        self.do_immediate_respawn = bool(compound["doimmediaterespawn"])
        self.do_insomnia = bool(compound["doinsomnia"])
        self.do_mob_loot = bool(compound["domobloot"])
        self.do_mob_spawning = bool(compound["domobspawning"])
        self.do_tile_drops = bool(compound["dotiledrops"])
        self.do_weather_cycle = bool(compound["doweathercycle"])
        self.drowning_damage = bool(compound["drowningdamage"])
        self.edu_level = bool(compound["eduLevel"])
        self.education_features_enabled = bool(compound["educationFeaturesEnabled"])
        self.experimental_game_play = bool(compound["experimentalgameplay"])
        self.fall_damage = bool(compound["falldamage"])
        self.fire_damage = bool(compound["firedamage"])
        self.function_command_limit = int(compound["functioncommandlimit"])
        self.has_been_loaded_in_creative = bool(compound["hasBeenLoadedInCreative"])
        self.has_locked_behaviour_pack = bool(compound["hasLockedBehaviorPack"])
        self.has_locked_resource_pack = bool(compound["hasLockedResourcePack"])
        self.immutable_world = bool(compound["immutableWorld"])
        self.is_created_in_editor = bool(compound["isCreatedInEditor"])
        self.is_exported_from_editor = bool(compound["isExportedFromEditor"])
        self.is_from_locked_template = bool(compound["isFromLockedTemplate"])
        self.is_from_world_template = bool(compound["isFromWorldTemplate"])
        self.is_world_template_option_locked = bool(
            compound["isWorldTemplateOptionLocked"]
        )
        self.keep_inventory = bool(compound["keepinventory"])
        self.last_opened_with_version = (
            int(compound["lastOpenedWithVersion"][0]),
            int(compound["lastOpenedWithVersion"][1]),
            int(compound["lastOpenedWithVersion"][2]),
            int(compound["lastOpenedWithVersion"][3]),
            int(compound["lastOpenedWithVersion"][4]),
        )
        self.lightning_level = float(compound["lightningLevel"])
        self.lightning_time = int(compound["lightningTime"])
        self.max_command_chain_length = int(compound["maxcommandchainlength"])
        self.mob_griefing = bool(compound["mobgriefing"])
        self.natural_regeneration = bool(compound["naturalregeneration"])
        self.prid = str(compound["prid"])
        self.pvp = bool(compound["pvp"])
        self.rain_level = float(compound["rainLevel"])
        self.rain_time = int(compound["rainTime"])
        self.random_tick_speed = int(compound["randomtickspeed"])
        self.requires_copied_pack_removal_check = bool(
            compound["requiresCopiedPackRemovalCheck"]
        )
        self.send_command_feedback = bool(compound["sendcommandfeedback"])
        self.server_chunk_tick_range = int(compound["serverChunkTickRange"])
        self.show_coordinates = bool(compound["showcoordinates"])
        self.show_death_messages = bool(compound["showdeathmessages"])
        self.spawn_mobs = bool(compound["spawnMobs"])
        self.spawn_radius = int(compound["spawnradius"])
        self.start_with_map_enabled = bool(compound["startWithMapEnabled"])
        self.texture_packs_required = bool(compound["texturePacksRequired"])
        self.tnt_explodes = bool(compound["tntexplodes"])
        self.use_msa_gamer_tags_only = bool(compound["useMsaGamertagsOnly"])
        self.world_start_count = int(compound["worldStartCount"])
        self.experiments = compound["experiments"]
        self.freeze_damage = bool(compound["freezedamage"])
        self.world_policies = compound["world_policies"]
        self.world_version = int(compound["WorldVersion"])
        self.respawn_blocks_explode = bool(compound["respawnblocksexplode"])
        self.show_border_effect = bool(compound["showbordereffect"])
        self.permissions_level = int(compound["permissionsLevel"])
        self.player_permissions_level = int(compound["playerPermissionsLevel"])
        self.is_random_seed_allowed = bool(compound["isRandomSeedAllowed"])
        self.do_limited_crafting = bool(compound["dolimitedcrafting"])
        self.editor_world_type = int(compound["editorWorldType"])
        self.players_sleeping_percentage = int(compound["playerssleepingpercentage"])
        self.recipes_unlock = bool(compound["recipesunlock"])
        self.natural_generation = bool(compound["naturalgeneration"])
        self.projectiles_can_break_blocks = bool(compound["projectilescanbreakblocks"])
        self.show_recipe_messages = bool(compound["showrecipemessages"])
        self.is_hardcore = bool(compound["IsHardcore"])
        self.show_days_played = bool(compound["showdaysplayed"])
        self.tnt_explosion_drop_decay = bool(compound["tntexplosiondropdecay"])
        self.has_uncomplete_world_file_on_disk = bool(
            compound["HasUncompleteWorldFileOnDisk"]
        )
        self.player_has_died = bool(compound["PlayerHasDied"])



import struct, nbtlib
from io import BytesIO
from .types import LIB
from .types import CSlice, CString, CInt, CLongLong
from .types import as_c_bytes, as_python_bytes, as_c_string, as_python_string
from ..utils import marshalNBT, unmarshalNBT


LIB.NewBedrockWorld.argtypes = [CString]
LIB.ReleaseBedrockWorld.argtypes = [CInt]
LIB.World_CloseWorld.argtypes = [CInt]
LIB.World_GetLevelDat.argtypes = [CInt]
LIB.World_ModifyLevelDat.argtypes = [CInt, CSlice]
LIB.LoadBiomes.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveBiomes.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.LoadChunkPayloadOnly.argtypes = [CInt, CInt, CInt, CInt]
LIB.LoadChunk.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveChunkPayloadOnly.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.SaveChunk.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.LoadSubChunk.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.SaveSubChunk.argtypes = [CInt, CInt, CInt, CInt, CInt, CInt]
LIB.LoadNBTPayloadOnly.argtypes = [CInt, CInt, CInt, CInt]
LIB.LoadNBT.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveNBTPayloadOnly.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.SaveNBT.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.LoadDeltaUpdate.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveDeltaUpdate.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.LoadTimeStamp.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveTimeStamp.argtypes = [CInt, CInt, CInt, CInt, CLongLong]
LIB.LoadDeltaUpdateTimeStamp.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveDeltaUpdateTimeStamp.argtypes = [CInt, CInt, CInt, CInt, CLongLong]
LIB.LoadFullSubChunkBlobHash.argtypes = [CInt, CInt, CInt, CInt]
LIB.SaveFullSubChunkBlobHash.argtypes = [CInt, CInt, CInt, CInt, CSlice]
LIB.LoadSubChunkBlobHash.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.SaveSubChunkBlobHash.argtypes = [CInt, CInt, CInt, CInt, CInt, CLongLong]

LIB.NewBedrockWorld.restype = CInt
LIB.ReleaseBedrockWorld.restype = None
LIB.World_CloseWorld.restype = CString
LIB.World_GetLevelDat.restype = CSlice
LIB.World_ModifyLevelDat.restype = CString
LIB.LoadBiomes.restype = CSlice
LIB.SaveBiomes.restype = CString
LIB.LoadChunkPayloadOnly.restype = CSlice
LIB.LoadChunk.restype = CInt
LIB.SaveChunkPayloadOnly.restype = CString
LIB.SaveChunk.restype = CString
LIB.LoadSubChunk.restype = CInt
LIB.SaveSubChunk.restype = CString
LIB.LoadNBTPayloadOnly.restype = CSlice
LIB.LoadNBT.restype = CSlice
LIB.SaveNBTPayloadOnly.restype = CString
LIB.SaveNBT.restype = CString
LIB.LoadDeltaUpdate.restype = CSlice
LIB.SaveDeltaUpdate.restype = CString
LIB.LoadTimeStamp.restype = CLongLong
LIB.SaveTimeStamp.restype = CString
LIB.LoadDeltaUpdateTimeStamp.restype = CLongLong
LIB.SaveDeltaUpdateTimeStamp.restype = CString
LIB.LoadFullSubChunkBlobHash.restype = CSlice
LIB.SaveFullSubChunkBlobHash.restype = CString
LIB.LoadSubChunkBlobHash.restype = CLongLong
LIB.SaveSubChunkBlobHash.restype = CString


def new_bedrock_world(dir: str) -> int:
    return int(LIB.NewBedrockWorld(as_c_string(dir)))


def release_bedrock_world(id: int) -> None:
    LIB.ReleaseBedrockWorld(CInt(id))


def world_close_world(id: int) -> str:
    return as_python_string(LIB.World_CloseWorld(CInt(id)))


def world_get_level_dat(id: int) -> tuple[nbtlib.tag.Compound | None, bool]:
    payload = as_python_bytes(LIB.World_GetLevelDat(CInt(id)))
    if len(payload) == 0:
        return (None, False)

    level_dat_data, _ = unmarshalNBT.UnMarshalBufferToPythonNBTObject(BytesIO(payload))
    return (level_dat_data, True)  # type: ignore


def world_modify_level_dat(id: int, level_dat: nbtlib.tag.Compound) -> str:
    writer = BytesIO()
    marshalNBT.MarshalPythonNBTObjectToWriter(writer, level_dat, "")
    return as_python_string(
        LIB.World_ModifyLevelDat(CInt(id), as_c_bytes(writer.getvalue()))
    )


def load_biomes(id: int, dm: int, x: int, z: int) -> bytes:
    return as_python_bytes(LIB.LoadBiomes(CInt(id), CInt(dm), CInt(x), CInt(z)))


def save_biomes(id: int, dm: int, x: int, z: int, payload: bytes) -> str:
    return as_python_string(
        LIB.SaveBiomes(CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(payload))
    )


def load_chunk_payload_only(id: int, dm: int, x: int, z: int) -> list[bytes]:
    payload = as_python_bytes(
        LIB.LoadChunkPayloadOnly(CInt(id), CInt(dm), CInt(x), CInt(z))
    )
    result = []

    ptr = 0
    while ptr < len(payload):
        l: int = struct.unpack("<I", payload[ptr : ptr + 4])[0]
        result.append(payload[ptr + 4 : ptr + 4 + l])
        ptr = ptr + 4 + l

    return result


def load_chunk(id: int, dm: int, x: int, z: int) -> int:
    return int(LIB.LoadChunk(CInt(id), CInt(dm), CInt(x), CInt(z)))


def save_chunk_payload_only(
    id: int, dm: int, x: int, z: int, payload: list[bytes]
) -> str:
    writer = BytesIO()

    for i in payload:
        l = struct.pack("<I", len(i))
        writer.write(l)
        writer.write(i)

    return as_python_string(
        LIB.SaveChunkPayloadOnly(
            CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(writer.getvalue())
        )
    )


def save_chunk(id: int, dm: int, x: int, z: int, chunk_id: int) -> str:
    return as_python_string(
        LIB.SaveChunk(CInt(id), CInt(dm), CInt(x), CInt(z), CInt(chunk_id))
    )


def load_sub_chunk(
    id: int,
    dm: int,
    x: int,
    y: int,
    z: int,
) -> int:
    return int(LIB.LoadSubChunk(CInt(id), CInt(dm), CInt(x), CInt(y), CInt(z)))


def save_sub_chunk(id: int, dm: int, x: int, y: int, z: int, sub_chunk_id: int) -> str:
    return as_python_string(
        LIB.SaveSubChunk(
            CInt(id), CInt(dm), CInt(x), CInt(y), CInt(z), CInt(sub_chunk_id)
        )
    )


def load_nbt_payload_only(id: int, dm: int, x: int, z: int) -> bytes:
    return as_python_bytes(LIB.LoadNBTPayloadOnly(CInt(id), CInt(dm), CInt(x), CInt(z)))


def load_nbt(id: int, dm: int, x: int, z: int) -> list[nbtlib.tag.Compound]:
    payload = as_python_bytes(LIB.LoadNBT(CInt(id), CInt(dm), CInt(x), CInt(z)))
    result = []

    ptr = 0
    while ptr < len(payload):
        l: int = struct.unpack("<I", payload[ptr : ptr + 4])[0]
        result.append(
            unmarshalNBT.UnMarshalBufferToPythonNBTObject(
                BytesIO(payload[ptr + 4 : ptr + 4 + l])
            )[0]
        )
        ptr = ptr + 4 + l

    return result


def save_nbt_payload_only(id: int, dm: int, x: int, z: int, payload: bytes) -> str:
    return as_python_string(
        LIB.SaveNBTPayloadOnly(
            CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(payload)
        )
    )


def save_nbt(id: int, dm: int, x: int, z: int, nbts: list[nbtlib.tag.Compound]) -> str:
    writer = BytesIO()

    for i in nbts:
        w = BytesIO()
        marshalNBT.MarshalPythonNBTObjectToWriter(w, i, "")

        binary_nbt = w.getvalue()
        l = struct.pack("<I", len(binary_nbt))

        writer.write(l)
        writer.write(binary_nbt)

    return as_python_string(
        LIB.SaveNBT(CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(writer.getvalue()))
    )


def load_delta_update(id: int, dm: int, x: int, z: int) -> bytes:
    return as_python_bytes(LIB.LoadDeltaUpdate(CInt(id), CInt(dm), CInt(x), CInt(z)))


def save_delta_update(id: int, dm: int, x: int, z: int, payload: bytes) -> str:
    return as_python_string(
        LIB.SaveDeltaUpdate(CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(payload))
    )


def load_time_stamp(id: int, dm: int, x: int, z: int) -> int:
    return int(LIB.LoadTimeStamp(CInt(id), CInt(dm), CInt(x), CInt(z)))


def save_time_stamp(id: int, dm: int, x: int, z: int, time_stamp: int) -> str:
    return as_python_string(
        LIB.SaveTimeStamp(CInt(id), CInt(dm), CInt(x), CInt(z), CLongLong(time_stamp))
    )


def load_delta_update_time_stamp(id: int, dm: int, x: int, z: int) -> int:
    return int(LIB.LoadDeltaUpdateTimeStamp(CInt(id), CInt(dm), CInt(x), CInt(z)))


def save_delta_update_time_stamp(
    id: int, dm: int, x: int, z: int, time_stamp: int
) -> str:
    return as_python_string(
        LIB.SaveDeltaUpdateTimeStamp(
            CInt(id), CInt(dm), CInt(x), CInt(z), CLongLong(time_stamp)
        )
    )


def load_full_sub_chunk_blob_hash(
    id: int, dm: int, x: int, z: int
) -> list[tuple[int, int]]:
    payload = as_python_bytes(
        LIB.LoadFullSubChunkBlobHash(CInt(id), CInt(dm), CInt(x), CInt(z))
    )
    result = []

    ptr = 0
    while ptr < len(payload):
        result.append(
            (payload[ptr], struct.unpack("<Q", payload[ptr + 1 : ptr + 9])[0])
        )
        ptr += 9

    return result


def save_full_sub_chunk_blob_hash(
    id: int, dm: int, x: int, z: int, hashes: list[tuple[int, int]]
) -> str:
    writer = BytesIO()

    for i in hashes:
        writer.write(i[0].to_bytes())
        writer.write(struct.pack("<Q", i[1]))

    return as_python_string(
        LIB.SaveFullSubChunkBlobHash(
            CInt(id), CInt(dm), CInt(x), CInt(z), as_c_bytes(writer.getvalue())
        )
    )


def load_sub_chunk_blob_hash(id: int, dm: int, x: int, y: int, z: int) -> int:
    return int(LIB.LoadSubChunkBlobHash(CInt(id), CInt(dm), CInt(x), CInt(y), CInt(z)))


def save_sub_chunk_blob_hash(
    id: int, dm: int, x: int, y: int, z: int, hash: int
) -> str:
    return as_python_string(
        LIB.SaveSubChunkBlobHash(
            CInt(id), CInt(dm), CInt(x), CInt(y), CInt(z), CLongLong(hash)
        )
    )



from dataclasses import dataclass, field
from .define import QuickSubChunkBlocks
from ..internal.symbol_export_sub_chunk import (
    new_sub_chunk as nsc,
    release_sub_chunk,
    sub_chunk_block,
    sub_chunk_blocks,
    sub_chunk_empty,
    sub_chunk_equals,
    sub_chunk_set_block,
    sub_chunk_set_blocks,
)


class SubChunkBase:
    """SubChunkBase is the base implement of a Minecraft sub chunk."""

    _sub_chunk_id: int

    def __init__(self):
        self._sub_chunk_id = -1

    def __del__(self):
        if self._sub_chunk_id >= 0 and not release_sub_chunk is None:
            release_sub_chunk(self._sub_chunk_id)

    def is_valid(self) -> bool:
        """
        is_valid check current sub chunk is valid or not.
        If not valid, it means the sub chunk is actually not exist, not only Python but also in Go.
        Try to use a invalid sub chunk is not allowed, and any operation will be terminated.

        Returns:
            bool: Wheatear the sub chunk is valid or not.
        """
        return self._sub_chunk_id >= 0


class SubChunk(SubChunkBase):
    """
    SubChunk is a cube of blocks located in a chunk.
    It has a size of 16x16x16 blocks and forms part of a stack that forms a Chunk.
    """

    def __init__(self):
        super().__init__()

    def equals(self, another_sub_chunk: SubChunkBase) -> bool:
        """Equals returns if the sub chunk passed is equal to the current one.

        Args:
            another_sub_chunk (SubChunkBase): The sub chunk passed.

        Returns:
            bool: The compare result.
                  True for the contents of two sub chunk is the same.
                  False for current sub chunk or another_sub_chunk is not found.
        """
        result = sub_chunk_equals(self._sub_chunk_id, another_sub_chunk._sub_chunk_id)
        return result == 1

    def empty(self) -> bool:
        """
        empty checks if the SubChunk is considered empty.
        This is the case if the SubChunk has 0 block storages or if it has a single one that is completely filled with air.

        Returns:
            bool: True for the sub chunk is empty.
                  False for the sub chunk is not empty, or the current sub chunk is not found.
        """
        result = sub_chunk_empty(self._sub_chunk_id)
        return result == 1

    def block(self, x: int, y: int, z: int, layer: int) -> int:
        """
        block returns the runtime ID of the block located at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.
            layer (int): The layer that the target block is in.

        Returns:
            int: Return the block runtime ID of target block.
                 If current sub chunk is not found, then return -1.
                 Note that if no sub chunk exists at the given y, the block is assumed to be air.
        """
        return sub_chunk_block(self._sub_chunk_id, x, y, z, layer)

    def blocks(self, layer: int) -> QuickSubChunkBlocks:
        """
        blocks all blocks (block runtime ids) whose in layer.
        It is highly suggest you use this instead of s.block(...) if you are trying to query
        so many blocks from this sub chunk.

        Args:
            layer (int): The sub chunk blocks you want to find.
                         layer refer to the finded blocks will from this layer.

        Returns:
            QuickSubChunkBlocks: All blocks of the target layer in this sub chunk if current sub chunk is exist.
                                 If the target layer is not exist, then you get a sub chunk full of air.
                                 Note that this implement don't do further check (maybe the underlying blocks list is empty)
                                 due to this is aims to increase block query/set speed, and you should take responsibility for
                                 any possible error.
        """
        return QuickSubChunkBlocks(sub_chunk_blocks(self._sub_chunk_id, layer))

    def set_block(self, x: int, y: int, z: int, layer: int, block_runtime_id: int):
        """
        set_block sets the given block runtime ID at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.
            layer (int): The layer that the target block is in.
            block_runtime_id (int): The block runtime id of target block will be.

        Raises:
            Exception: When failed to set block.
        """
        err = sub_chunk_set_block(self._sub_chunk_id, x, y, z, layer, block_runtime_id)
        if len(err) > 0:
            raise Exception(err)

    def set_blocks(self, layer: int, blocks: QuickSubChunkBlocks):
        """
        set_blocks sets the whole chunk blocks in layer by given block runtime ids.

        It is highly suggest you use this instead of s.set_block(...) if you are trying to modify
        so many blocks to this sub chunk.

        Note that this implement will not check the underlying blocks list is valid
        or not due to this is aims to increase block query/set speed, and you should
        take responsibility for any possible error.

        Args:
            layer (int): The blocks in the target layer of this chunk that you want to overwrite.

        Raises:
            Exception: When failed to set blocks.
        """
        err = sub_chunk_set_blocks(self._sub_chunk_id, layer, blocks.blocks)
        if len(err) > 0:
            raise Exception(err)


@dataclass
class SubChunkWithIndex:
    """SubChunkWithIndex represents a sub chunk and its index in a whole chunk where this sub chunk is in.

    Args:
        sub_chunk (SubChunk): The sub chunk.
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0
    """

    index: int = 0
    sub_chunk: SubChunk = field(default_factory=lambda: SubChunk())


def new_sub_chunk() -> SubChunk:
    """
    NewSubChunk creates a new sub chunk.
    All sub chunks should be created through this function.

    Returns:
        SubChunk: A new sub chunk that full of air.
    """
    s = SubChunk()
    s._sub_chunk_id = nsc()
    return s



import nbtlib, numpy
from dataclasses import dataclass, field


@dataclass
class ChunkPos:
    """
    ChunkPos holds the position of a chunk. The type is provided as a utility struct for keeping track of a
    chunk's position. Chunks do not themselves keep track of that. Chunk positions are different from block
    positions in the way that increasing the X/Z by one means increasing the absolute value on the X/Z axis in
    terms of blocks by 16.
    """

    x: int = 0
    z: int = 0


@dataclass
class SubChunkPos:
    """
    SubChunkPos holds the position of a sub-chunk. The type is provided as a utility struct for keeping track of a
    sub-chunk's position. Sub-chunks do not themselves keep track of that. Sub-chunk positions are different from
    block positions in the way that increasing the X/Y/Z by one means increasing the absolute value on the X/Y/Z axis in
    terms of blocks by 16.
    """

    x: int = 0
    y: int = 0
    z: int = 0


@dataclass
class Range:
    """
    Range represents the height range of a Dimension in blocks. The first value
    of the Range holds the minimum Y value, the second value holds the maximum Y
    value.
    """

    start_range: int = 0
    end_range: int = 0


class Dimension:
    """
    Dimension is a dimension of a World. It influences a variety of
    properties of a World such as the building range, the sky colour and the
    behaviour of liquid blocks.
    """

    dm: int = 0

    def __init__(self, dm: int):
        """Init a new dimension represent.

        Args:
            dm (int): The id of this dimension.
        """
        self.dm = dm

    def range(self) -> Range:
        """range return the range that player could build block in this dimension.

        Returns:
            Range: The range that player could build block in this dimension.
                   If this dimension is not standard dimension, then redirect
                   to overworld range.
        """
        match self.dm:
            case 0:
                return Range(-64, 319)
            case 1:
                return Range(0, 127)
            case 2:
                return Range(0, 255)
            case _:
                return Range(-64, 319)

    def height(self) -> int:
        """
        height returns the height of this dimension.
        For example, the height of overworld is 384
        due to "384 = 319 - (-64) + 1", and 319 is
        the max Y that overworld could build, and -64
        is the min Y that overworld could build.

        Returns:
            int: The height of this dimension.
                 If this dimension is not standard dimension, then redirect
                 to overworld height.
        """
        match self.dm:
            case 0:
                return 384
            case 1:
                return 128
            case 2:
                return 256
            case _:
                return 384

    def __str__(self) -> str:
        match self.dm:
            case 0:
                return "Overworld"
            case 1:
                return "Nether"
            case 2:
                return "End"
            case _:
                return f"Custom (id={self.dm})"


@dataclass
class BlockStates:
    """BlockState holds a combination of a name and properties."""

    Name: str = ""
    States: nbtlib.tag.Compound = field(default_factory=lambda: nbtlib.tag.Compound())


# ptr = ((y >> 4) - (self.start_range >> 4)) << 12
# offset = x * 256 + (y & 15) * 16 + z
@dataclass
class QuickChunkBlocks:
    """
    QuickChunkBlocks is a quick blocks getter and setter, which used for a Minecraft chunk.
    Note that it is only represent one layer in this chunk.

    Args:
        blocks (list[int], optional): A dense matrix that represent each block in a layer of this chunk.
                                      Defaults to empty list.
        start_range (int): The min Y position of this chunk.
                           For overworld is -64, but nether and end is 0.
                           Defaults to -64.
        end_range (int): The max Y position of this chunk.
                         For overworld is 319, for nether is 127, and for end is 255.
                         Defaults to 319.
    """

    blocks: numpy.ndarray = field(
        default_factory=lambda: numpy.array([], dtype=numpy.uint32)
    )
    start_range: int = -64
    end_range: int = 319

    def set_empty(self, air_block_runtime_id: int):
        """set_empty make this chunk full of air.

        Args:
            air_block_runtime_id (int): The block runtime id of air block.
        """
        self.blocks = numpy.full(
            4096 * ((self.end_range - self.start_range + 1) >> 4),
            air_block_runtime_id,
            dtype=numpy.uint32,
        )

    def block(self, x: int, y: int, z: int) -> numpy.uint32:
        """Block returns the runtime ID of the block at a given x, y and z in this chunk.

        Args:
            x (int): The relative x position of this block. Must in a range of 0-15.
            y (int): The y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this block. Must in a range of 0-15.

        Returns:
            int: Return the block runtime ID of target block.
                 It will not check whether the index is overflowing.
        """
        return self.blocks[
            (((y >> 4) - (self.start_range >> 4)) << 12) + x * 256 + (y & 15) * 16 + z
        ]

    def set_block(self, x: int, y: int, z: int, block_runtime_id: int | numpy.uint32):
        """
        set_block sets the runtime ID of a block at a given x, y and z in this chunk.
        Note that:
            - This operation is just on program, and you need to use c.set_blocks(layer, QuickChunkBlocks) to apply
              changes to the chunk. Then, after you apply changes, use w.save_chunk(...) to apply changes to the game saves.
            - It will not check whether the index is overflowing.

        Args:
            x (int): The relative x position of this block. Must in a range of 0-15.
            y (int): The y position of this block.
                     Must in a range of -64~319 (Overworld), 0-127 (Nether) and 0-255 (End).
            z (int): The relative z position of this block. Must in a range of 0-15.
            block_runtime_id (int): The result block that this block will be.
        """
        self.blocks[
            (((y >> 4) - (self.start_range >> 4)) << 12) + x * 256 + (y & 15) * 16 + z
        ] = block_runtime_id


@dataclass
class QuickSubChunkBlocks:
    """
    QuickSubChunkBlocks is a quick blocks getter and setter, which used for a Minecraft sub chunk.
    Note that it is only represent one layer in this sub chunk.

    Args:
        blocks (list[int], optional): A dense matrix that represent each block in a layer of this sub chunk.
                                      Defaults to empty list.
    """

    blocks: numpy.ndarray = field(
        default_factory=lambda: numpy.array([], dtype=numpy.uint32)
    )

    def set_empty(self, air_block_runtime_id: int):
        """set_empty make this sub chunk full of air.

        Args:
            air_block_runtime_id (int): The block runtime id of air block.
        """
        self.blocks = numpy.full(4096, air_block_runtime_id, dtype=numpy.uint32)

    def block(self, x: int, y: int, z: int) -> numpy.uint32:
        """
        block returns the runtime ID of the block located at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.

        Returns:
            int: Return the block runtime ID of target block.
                 It will not check whether the index is overflowing.
        """
        return self.blocks[x * 256 + y * 16 + z]

    def set_block(self, x: int, y: int, z: int, block_runtime_id: int | numpy.uint32):
        """
        set_block sets the given block runtime ID at the given X, Y and Z.
        X, Y and Z must be in a range of 0-15.
        Note that:
            - This operation is just on program, and you need to use s.set_blocks(layer, QuickSubChunkBlocks) to apply changes
              to the sub chunk. Then, after you apply changes,
                - use w.save_sub_chunk(...) to apply changes to the game saves.
                - if this sub chunk is from a loaded chunk, then you'd be suggest to use w.save_chunk(...) to apply changes
                  to the game saves if there are multiple sub chunk changes in the target chunk.
            - It will not check whether the index is overflowing.

        Args:
            x (int): The relative x position of target block. Must in a range of 0-15.
            y (int): The relative y position of target block. Must in a range of 0-15.
            z (int): The relative z position of target block. Must in a range of 0-15.
            block_runtime_id (int): The block runtime id of target block will be.
        """
        self.blocks[x * 256 + y * 16 + z] = block_runtime_id


@dataclass
class HashWithPosY:
    Hash: int = 0
    PosY: int = 0



from .world.chunk import Chunk, new_chunk
from .world.sub_chunk import SubChunk, SubChunkWithIndex, new_sub_chunk
from .world.world import World, new_world
from .world.level_dat import LevelDat, Abilities

from .world.constant import (
    EMPTY_COMPOUND,
    EMPTY_BLOCK_STATES,
    DIMENSION_ID_OVERWORLD,
    DIMENSION_ID_NETHER,
    DIMENSION_ID_END,
    DIMENSION_OVERWORLD,
    DIMENSION_NETHER,
    DIMENSION_END,
    RANGE_OVERWORLD,
    RANGE_NETHER,
    RANGE_END,
    RANGE_INVALID,
    AIR_BLOCK_STATES,
    AIR_BLOCK_RUNTIME_ID,
)

from .world.define import (
    ChunkPos,
    SubChunkPos,
    Range,
    Dimension,
    BlockStates,
    QuickChunkBlocks,
    QuickSubChunkBlocks,
    HashWithPosY,
)

from .world.conversion import (
    runtime_id_to_state,
    state_to_runtime_id,
    sub_chunk_network_payload,
    from_sub_chunk_network_payload,
    sub_chunk_disk_payload,
    from_sub_chunk_disk_payload,
)

from nbtlib.tag import Compound, String, Byte, Int



import os, ctypes, platform
from pathlib import Path

LIB = None

package_dir = Path(__file__).parent.parent.resolve()
lib_path = os.path.join(package_dir, "dynamic_libs")

system = platform.system().lower()
arch = platform.machine().lower()

match system:
    case "windows":
        if arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_windows_amd64.dll")
            )
        elif arch == "x86_64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_windows_x86.dll")
            )
    case "darwin":
        if arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_macos_amd64.dylib")
            )
        elif arch == "arm64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_macos_arm64.dylib")
            )
    case _:
        if arch == "aarch64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_android_arm64.so")
            )
        elif arch == "amd64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_linux_amd64.so")
            )
        elif arch == "arm64":
            LIB = ctypes.cdll.LoadLibrary(
                os.path.join(lib_path, "bedrock-world-operator_linux_arm64.so")
            )

if LIB is None:
    raise FileNotFoundError("Your machine is not supported.")



import numpy
from .types import LIB, CSlice, CString, as_c_bytes, as_python_bytes, as_python_string
from .types import CInt


LIB.NewSubChunk.argtypes = []
LIB.ReleaseSubChunk.argtypes = [CInt]
LIB.SubChunk_Block.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.SubChunk_Blocks.argtypes = [CInt, CInt]
LIB.SubChunk_Empty.argtypes = [CInt]
LIB.SubChunk_Equals.argtypes = [CInt, CInt]
LIB.SubChunk_SetBlock.argtypes = [CInt, CInt, CInt, CInt, CInt, CInt]
LIB.SubChunk_SetBlocks.argtypes = [CInt, CInt, CSlice]

LIB.NewSubChunk.restype = CInt
LIB.ReleaseSubChunk.restype = None
LIB.SubChunk_Block.restype = CInt
LIB.SubChunk_Blocks.restype = CSlice
LIB.SubChunk_Empty.restype = CInt
LIB.SubChunk_Equals.restype = CInt
LIB.SubChunk_SetBlock.restype = CString
LIB.SubChunk_SetBlocks.restype = CString


def new_sub_chunk() -> int:
    return int(LIB.NewSubChunk())


def release_sub_chunk(id: int) -> None:
    LIB.ReleaseSubChunk(CInt(id))


def sub_chunk_block(id: int, x: int, y: int, z: int, layer: int) -> int:
    return int(LIB.SubChunk_Block(CInt(id), CInt(x), CInt(y), CInt(z), CInt(layer)))


def sub_chunk_blocks(id: int, layer: int) -> numpy.ndarray:
    return numpy.frombuffer(
        as_python_bytes(LIB.SubChunk_Blocks(CInt(id), CInt(layer))), dtype=numpy.uint32
    ).copy()


def sub_chunk_empty(id: int) -> int:
    return int(LIB.SubChunk_Empty(CInt(id)))


def sub_chunk_equals(id: int, another_sub_chunk_id: int) -> int:
    return int(LIB.SubChunk_Equals(CInt(id), CInt(another_sub_chunk_id)))


def sub_chunk_set_block(
    id: int, x: int, y: int, z: int, layer: int, block_runtime_id: int
) -> str:
    return as_python_string(
        LIB.SubChunk_SetBlock(
            CInt(id), CInt(x), CInt(y), CInt(z), CInt(layer), CInt(block_runtime_id)
        )
    )


def sub_chunk_set_blocks(id: int, layer: int, blocks: numpy.ndarray) -> str:
    return as_python_string(
        LIB.SubChunk_SetBlocks(CInt(id), CInt(layer), as_c_bytes(blocks.tobytes()))
    )



from .types import LIB
from .types import CSlice, CString, CInt
from .types import as_c_bytes, as_python_bytes, as_python_string


LIB.DB_Has.argtypes = [CInt, CSlice]
LIB.DB_Get.argtypes = [CInt, CSlice]
LIB.DB_Put.argtypes = [CInt, CSlice, CSlice]
LIB.DB_Delete.argtypes = [CInt, CSlice]

LIB.DB_Has.restype = CInt
LIB.DB_Get.restype = CSlice
LIB.DB_Put.restype = CString
LIB.DB_Delete.restype = CString


def db_has(world_id: int, key: bytes) -> int:
    return int(LIB.DB_Has(CInt(world_id), as_c_bytes(key)))


def db_get(world_id: int, key: bytes) -> bytes:
    return as_python_bytes(LIB.DB_Get(CInt(world_id), as_c_bytes(key)))


def db_put(world_id: int, key: bytes, value: bytes) -> str:
    return as_python_string(
        LIB.DB_Put(CInt(world_id), as_c_bytes(key), as_c_bytes(value))
    )


def db_delete(world_id: int, key: bytes) -> str:
    return as_python_string(LIB.DB_Delete(CInt(world_id), as_c_bytes(key)))



from io import BytesIO
from struct import pack
import nbtlib

endian: str = '<'
endianWord: str = 'little'


def getValueType(value) -> bytes:
    match type(value):
        case nbtlib.tag.Byte:
            return b'\x01'
        case nbtlib.tag.Short:
            return b'\x02'
        case nbtlib.tag.Int:
            return b'\x03'
        case nbtlib.tag.Long:
            return b'\x04'
        case nbtlib.tag.Float:
            return b'\x05'
        case nbtlib.tag.Double:
            return b'\x06'
        case nbtlib.tag.ByteArray:
            return b'\x07'
        case nbtlib.tag.String:
            return b'\x08'
        case nbtlib.tag.Compound:
            return b'\x0a'
        case nbtlib.tag.IntArray:
            return b'\x0b'
        case nbtlib.tag.LongArray:
            return b'\x0c'
        case _:
            return b'\x09'


def marshalToName(writer: BytesIO, name: str) -> None:
    encodeResult = name.encode(encoding='utf-8')
    writer.write(pack(f'{endian}H', len(encodeResult)) + encodeResult)


def marshalToValue(writer: BytesIO, value, valueType: int) -> None:
    match valueType:
        case 1:
            writer.write(value.to_bytes(length=1, byteorder=endianWord, signed=True))
        case 2:
            writer.write(pack(f'{endian}h', value))
        case 3:
            writer.write(pack(f'{endian}i', value))
        case 4:
            writer.write(pack(f'{endian}q', value))
        case 5:
            writer.write(pack(f'{endian}f', value))
        case 6:
            writer.write(pack(f'{endian}d', value))
        case 7 | 11 | 12:
            marshalToArray(writer, value, valueType)
        case 8:
            marshalToName(writer, str(value))
        case 9:
            marshalToList(writer, value)
        case 10:
            marshalToCompound(writer, value)


def marshalToArray(
        writer: BytesIO,
        value: nbtlib.tag.ByteArray | nbtlib.tag.IntArray | nbtlib.tag.LongArray,
        valueType: int
) -> None:
    writer.write(pack(f'{endian}i', len(value)))
    match valueType:
        case 7:
            writer.write(b''.join([pack(f'{endian}b', i) for i in value]))
        case 11:
            writer.write(b''.join([pack(f'{endian}i', i) for i in value]))
        case 12:
            writer.write(b''.join([pack(f'{endian}q', i) for i in value]))
        case _:
            raise TypeError


def marshalToList(writer: BytesIO, value) -> None:
    if len(value) > 0:
        subValueType = getValueType(value[0])
        writer.write(subValueType+pack(f'{endian}i', len(value)))
        for i in value:
            marshalToValue(writer, i, subValueType[0])
    else:
        writer.write(b'\x00' * 5)


def marshalToCompound(writer: BytesIO, value: nbtlib.tag.Compound) -> None:
    for i in value:
        valueType = getValueType(value[i])
        writer.write(valueType)
        marshalToName(writer, i)
        marshalToValue(writer, value[i], valueType[0])
    writer.write(b'\x00')


def MarshalPythonNBTObjectToWriter(writer: BytesIO, value, name: str) -> None:
    valueType = getValueType(value)
    writer.write(valueType)
    marshalToName(writer, name)
    marshalToValue(writer, value, valueType[0])



import nbtlib
from .constant import DIMENSION_OVERWORLD
from ..world.chunk import Chunk
from ..world.sub_chunk import SubChunk
from ..internal.symbol_export_world import (
    load_biomes,
    load_chunk,
    load_chunk_payload_only,
    load_delta_update,
    load_delta_update_time_stamp,
    load_full_sub_chunk_blob_hash,
    load_nbt,
    load_nbt_payload_only,
    load_sub_chunk,
    load_sub_chunk_blob_hash,
    load_time_stamp,
    new_bedrock_world as nbw,
    release_bedrock_world,
    save_biomes,
    save_chunk,
    save_chunk_payload_only,
    save_delta_update,
    save_delta_update_time_stamp,
    save_full_sub_chunk_blob_hash,
    save_nbt,
    save_nbt_payload_only,
    save_sub_chunk,
    save_sub_chunk_blob_hash,
    save_time_stamp,
    world_close_world,
    world_get_level_dat,
    world_modify_level_dat,
)
from ..internal.symbol_export_world_underlying import (
    db_delete,
    db_get,
    db_has,
    db_put,
)
from ..world.define import ChunkPos, Dimension, HashWithPosY, SubChunkPos
from ..world.level_dat import LevelDat


class WorldBase:
    """WorldBase is the base implement of a Minecraft world game saves."""

    _world_id: int

    def __init__(self):
        self._world_id = -1

    def __del__(self):
        if self._world_id >= 0 and not release_bedrock_world is None:
            release_bedrock_world(self._world_id)

    def is_valid(self) -> bool:
        """
        is_valid check the opened world is valid or not.
        If not valid, it means current world is actually not opened.
        Try to use a world that not opened is not allowed, and any operation will be terminated.

        Returns:
            bool: The current world is valid or not.
        """
        return self._world_id >= 0

    def close_world(self):
        """close_world close this game saves.

        Raises:
            Exception: When failed to close the world.
        """
        err = world_close_world(self._world_id)
        if len(err) > 0:
            raise Exception(err)

    def has(self, key: bytes) -> bool:
        """has check the if the key is in the underlying database of this game saves.

        Args:
            key (bytes): The bytes represent of the key.

        Returns:
            bool: If key is exist, then this is true.
                  Otherwise, when meet error or not exist, return false.
        """
        return db_has(self._world_id, key) == 1

    def get(self, key: bytes) -> bytes:
        """get try to get the value of the key in the underlying database.

        Args:
            key (bytes): The bytes represent of the key.

        Returns:
            bytes: If key is exist, then return the value of this key.
                   Otherwise, when meet error or not exist, return empty bytes.
        """
        return db_get(self._world_id, key)

    def put(self, key: bytes, value: bytes):
        """put set the key to value in the underlying database.

        Args:
            key (bytes): The bytes represent of the key.
            value (bytes): The bytes represent of the value.

        Raises:
            Exception: When failed to set the value of this key.
        """
        err = db_put(self._world_id, key, value)
        if len(err) > 0:
            raise Exception(err)

    def delete(self, key: bytes):
        """delete remove the key and its value from the underlying database.

        Args:
            key (bytes): The bytes represent of the key.

        Raises:
            Exception: When failed to remove the key from the database.
        """
        err = db_delete(self._world_id, key)
        if len(err) > 0:
            raise Exception(err)


class World(WorldBase):
    """
    World is the completely implements of Minecraft bedrock game saves,
    which only entities and player data related things are not implement.
    """

    def __init__(self):
        super().__init__()

    def get_level_dat(self) -> LevelDat | None:
        """get_level_dat get the level dat of current game saves.

        Returns:
            LevelDat | None:
                If success, then return the level dat.
                Otherwise, return None.
        """
        ldt = LevelDat()

        result, success = world_get_level_dat(self._world_id)
        if not success:
            return None

        ldt.unmarshal(result)  # type: ignore
        return ldt

    def modify_level_dat(self, new_level_dat: LevelDat):
        """modify_level_dat set the level dat of this world to new_level_dat.

        Args:
            new_level_dat (LevelDat): The new level dat want to set.

        Raises:
            Exception: When failed to set level dat.
        """
        err = world_modify_level_dat(self._world_id, new_level_dat.marshal())
        if len(err) > 0:
            raise Exception(err)

    def load_biomes(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> bytes:
        """load_biomes loads the biome data of a chunk whose in chunk_pos and dm.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            bytes: The biome data of target chunk.
                   If meet error or not exist, then return empty bytes.
        """
        return load_biomes(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def save_biomes(
        self,
        chunk_pos: ChunkPos,
        biomes_data: bytes,
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """save_biomes set the biome data of a chunk whose in chunk_pos and dm.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            biomes_data (bytes): The biome data want to set to this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to set biome data.
        """
        err = save_biomes(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, biomes_data)
        if len(err) > 0:
            raise Exception(err)

    def load_chunk_payload_only(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> list[bytes]:
        """
        load_chunk_payload_only loads a chunk at the position passed from the leveldb database.
        Note that we here don't decode chunk data and just return the origin payload.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            list[bytes]: The raw payload of this chunk, where each element in the list corresponds to a sub block payload.
                         If meet error or not exist, then return empty list.
        """
        return load_chunk_payload_only(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def load_chunk(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> Chunk:
        """load_chunk loads a chunk at the position passed from the leveldb database.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            Chunk: If current world or the chunk is not exist, then return a invalid chunk.
                   Otherwise, return the target chunk and True.
                   Note that you could use c.is_valid() to check whether the chunk is valid or not.
        """
        c = Chunk()
        c._chunk_id = load_chunk(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)
        return c

    def save_chunk_payload_only(
        self,
        chunk_pos: ChunkPos,
        payload: list[bytes],
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """
        save_chunk_payload_only saves multiple sub chunks payload to the leveldb database which all these sub chunks are in a chunk.
        Note that we also write the version of this chunk.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            payload (list[bytes]): The payload of each sub chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save payload chunk.
        """
        err = save_chunk_payload_only(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, payload
        )
        if len(err) > 0:
            raise Exception(err)

    def save_chunk(
        self,
        chunk_pos: ChunkPos,
        chunk: Chunk,
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """
        save_chunk saves a chunk at the position passed to the leveldb database.
        Note that we also write the version of this chunk.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            chunk (Chunk): The chunk we want to save to the game saves.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save chunk.
        """
        err = save_chunk(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, chunk._chunk_id
        )
        if len(err) > 0:
            raise Exception(err)

    def load_sub_chunk(
        self,
        sub_chunk_pos: SubChunkPos,
        dm: Dimension = DIMENSION_OVERWORLD,
    ) -> SubChunk:
        """load_sub_chunk loads a sub chunk at the position from the leveldb database.

        Args:
            sub_chunk_pos (SubChunkPos): The sub chunk pos of this sub chunk.
            dm (Dimension, optional): The dimension of this sub chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            SubChunk: If current world or the sub chunk is not exist, then return a invalid sub chunk.
                      Otherwise, return the target sub chunk and True.
                      Note that you could use s.is_valid() to check whether the sub chunk is valid or not.
        """
        s = SubChunk()
        s._sub_chunk_id = load_sub_chunk(
            self._world_id, dm.dm, sub_chunk_pos.x, sub_chunk_pos.y, sub_chunk_pos.z
        )
        return s

    def save_sub_chunk(
        self,
        sub_chunk_pos: SubChunkPos,
        sub_chunk: SubChunk,
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """
        save_sub_chunk saves a sub chunk at the position passed to the leveldb database.
        Note that we also write the version of the whole chunk where this sub chunk is in.

        Args:
            sub_chunk_pos (SubChunkPos): The sub chunk pos of this sub chunk.
            sub_chunk (SubChunk): The sub chunk we want to save to the game saves.
            dm (Dimension, optional): The dimension of this sub chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save sub chunk.
        """
        err = save_sub_chunk(
            self._world_id,
            dm.dm,
            sub_chunk_pos.x,
            sub_chunk_pos.y,
            sub_chunk_pos.z,
            sub_chunk._sub_chunk_id,
        )
        if len(err) > 0:
            raise Exception(err)

    def load_nbt_payload_only(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> bytes:
        """
        load_nbt_payload_only loads payload of all block entities from the chunk position passed.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            bytes: The raw NBT payload of all block entities in this chunk.
                   If meet error or not exist, then return empty bytes.
        """
        return load_nbt_payload_only(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def load_nbt(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> list[nbtlib.tag.Compound]:
        """load_nbt loads all block entities from the chunk position passed.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            list[nbtlib.tag.Compound]: The decoded block entities NBT of this chunk.
                                       If meet error or not exist, then return empty list.
        """
        return load_nbt(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def save_nbt_payload_only(
        self, chunk_pos: ChunkPos, payload: bytes, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """save_nbt_payload_only saves a serialized NBT data to the chunk position passed.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            payload (bytes): The raw payload of all block entities NBT in this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save payload NBT
        """
        err = save_nbt_payload_only(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, payload
        )
        if len(err) > 0:
            raise Exception(err)

    def save_nbt(
        self,
        chunk_pos: ChunkPos,
        nbts: list[nbtlib.tag.Compound],
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """save_nbt saves all block NBT data to the chunk position passed.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            nbts (list[nbtlib.tag.Compound]): A list holds all block entities NBT data of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save NBT
        """
        err = save_nbt(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, nbts)
        if len(err) > 0:
            raise Exception(err)

    def load_delta_update(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> bytes:
        """
        load_delta_update load a custom purpose payload which related a chunk from leveldb database.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            bytes: The delta update payload of this chunk.
                   If meet error or not exist, then return empty bytes.
        """
        return load_delta_update(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def save_delta_update(
        self, chunk_pos: ChunkPos, payload: bytes, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """
        save_delta_update save a custom purpose payload which related a chunk to leveldb database.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            dm (Dimension): The dimension of this chunk.
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            payload (bytes): The delta update payload want to set for this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save delta update payload.
        """
        err = save_delta_update(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, payload
        )
        if len(err) > 0:
            raise Exception(err)

    def load_time_stamp(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> int:
        """
        load_time_stamp load the last update unix time of a chunk whose in chunk_pos and dm.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            int: The last update unix time of this chunk.
                 Return -1 for current world is not exist,
                 Return 0 for the time stamp is not exist.
        """
        return load_time_stamp(self._world_id, dm.dm, chunk_pos.x, chunk_pos.z)

    def save_time_stamp(
        self, chunk_pos: ChunkPos, time_stamp: int, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """
        save_time_stamp save the last update unix time of a chunk whose in chunk_pos and dm.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            time_stamp (int): The last update unix time that want to set for this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save time stamp.
        """
        err = save_time_stamp(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, time_stamp
        )
        if len(err) > 0:
            raise Exception(err)

    def load_delta_time_stamp(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> int:
        """
        load_delta_time_stamp load the last update unix time of a delta update
        which related to a chunk in chunk_pos and dm.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of the chunk.
            dm (Dimension, optional): The dimension of the chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            int: The last update unix time of the delta update.
                 Return -1 for current world is not exist,
                 Return 0 for the time stamp is not exist.
        """
        return load_delta_update_time_stamp(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z
        )

    def save_delta_time_stamp(
        self, chunk_pos: ChunkPos, time_stamp: int, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """
        save_time_stamp save the last update unix time of a delta update which related to chunk in chunk_pos and dm.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of the chunk.
            time_stamp (int): The last update unix time that want to set for the delta update.
            dm (Dimension, optional): The dimension of the chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save time stamp.
        """
        err = save_delta_update_time_stamp(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z, time_stamp
        )
        if len(err) > 0:
            raise Exception(err)

    def load_full_sub_chunk_blob_hash(
        self, chunk_pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ) -> list[HashWithPosY]:
        """
        load_full_sub_chunk_blob_hash loads the blob hash of a chunk.
        Actually speaking, this is hash for multiple sub chunks
        (each sub chunks who have payload will have a hash value), not just a completely chunk.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            list[HashWithPosY]: The hashes of this chunk.
                                If not exist or meet error, then return empty list.
        """
        hashes = load_full_sub_chunk_blob_hash(
            self._world_id, dm.dm, chunk_pos.x, chunk_pos.z
        )
        result = []
        for i in hashes:
            result.append(HashWithPosY(i[1], i[0]))
        return result

    def save_full_sub_chunk_blob_hash(
        self,
        chunk_pos: ChunkPos,
        new_hash: list[HashWithPosY],
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """
        save_full_sub_chunk_blob_hash update the blob hash of a chunk.
        It's necessary to say that this is not a hash of a completely chunk,
        but are multiple hash for each sub chunk who have payload in this chunk.
        Note that:
            - If len(new_hash) is 0, then the blob hash
              data of this chunk will be delete.
            - Zero hash is allowed.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            chunk_pos (ChunkPos): The chunk pos of this chunk.
            hashes (list[HashWithPosY]): The hashes want to save to this chunk.
            dm (Dimension, optional): The dimension of this chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save blob hash of this chunk.
        """
        err = save_full_sub_chunk_blob_hash(
            self._world_id,
            dm.dm,
            chunk_pos.x,
            chunk_pos.z,
            [(i.PosY, i.Hash) for i in new_hash],
        )
        if len(err) > 0:
            raise Exception(err)

    def load_sub_chunk_blob_hash(
        self,
        sub_chunk_pos: SubChunkPos,
        dm: Dimension = DIMENSION_OVERWORLD,
    ) -> int:
        """
        load_sub_chunk_blob_hash loads the blob hash of a sub chunk that in sub_chunk_pos and in dm dimension.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            sub_chunk_pos (SubChunkPos): The sub chunk pos of this sub chunk.
            dm (Dimension, optional): The dimension of this sub chunk. Defaults to DIMENSION_OVERWORLD.

        Returns:
            int: The blob hash value of target sub chunk.
                 If meet error or not exist, then return -1.
                 Return 0 for a sub chunk that full of air.
        """
        return load_sub_chunk_blob_hash(
            self._world_id, dm.dm, sub_chunk_pos.x, sub_chunk_pos.y, sub_chunk_pos.z
        )

    def save_sub_chunk_blob_hash(
        self,
        sub_chunk_pos: SubChunkPos,
        hash: int,
        dm: Dimension = DIMENSION_OVERWORLD,
    ):
        """
        save_sub_chunk_blob_hash save the hash for sub chunk which in sub_chunk_pos and in dm dimension.
        Note that zero hash is allowed.
        This is not used by standard Minecraft, and just for some own purpose.

        Args:
            sub_chunk_pos (SubChunkPos): The sub chunk pos of this sub chunk.
            dm (Dimension, optional): The dimension of this sub chunk. Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to save blob hash of this sub chunk.
        """
        err = save_sub_chunk_blob_hash(
            self._world_id,
            dm.dm,
            sub_chunk_pos.x,
            sub_chunk_pos.y,
            sub_chunk_pos.z,
            hash,
        )
        if len(err) > 0:
            raise Exception(err)


def new_world(dir: str) -> World:
    """
    new_world creates a new provider reading and writing from/to files under the path
    passed using default options. If a world is present at the path, new_world will
    parse its data and initialise the world with it.

    Args:
        dir (str): The minecraft bedrock leveldb path (folder path)

    Returns:
        World: If database can't be initialise or the level dat is cannot be
               parsed, then return a invalid world.
               Otherwise, return the bedrock world and True.
               Note that you could use w.is_valid() to check the world is valid or not.
    """
    w = World()
    w._world_id = nbw(dir)
    return w



from .load_dynamic_library import LIB
import ctypes, struct

CPtr = ctypes.c_void_p
CSlice = CPtr
CString = CPtr
CInt = ctypes.c_int
CLongLong = ctypes.c_longlong


LIB.FreeMemory.argtypes = [CPtr]
LIB.FreeMemory.restype = None


def free_memory(address: CPtr) -> None:
    LIB.FreeMemory(address)


def as_c_bytes(b: bytes) -> CSlice:
    return ctypes.cast(ctypes.c_char_p(struct.pack("<I", len(b)) + b), CSlice)


def as_python_bytes(slice: CSlice) -> bytes:
    l = struct.unpack("<I", ctypes.string_at(slice, 4))[0]
    result = ctypes.string_at(slice, 4 + l)[4:]
    free_memory(slice)
    return result


def as_c_string(string: str) -> CString:
    return ctypes.cast(ctypes.c_char_p(bytes(string, encoding="utf-8")), CString)


def as_python_string(c_string: CString) -> str:
    result = ctypes.c_char_p(c_string).value.decode(encoding="utf-8")  # type: ignore
    free_memory(c_string)
    return result



import struct, numpy
from .types import LIB
from .types import CSlice, CString, CInt
from .types import as_c_bytes, as_python_bytes, as_python_string


LIB.NewChunk.argtypes = [CInt, CInt]
LIB.ReleaseChunk.argtypes = [CInt]
LIB.Chunk_Biome.argtypes = [CInt, CInt, CInt, CInt]
LIB.Chunk_Block.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.Chunk_Blocks.argtypes = [CInt, CInt]
LIB.Chunk_Compact.argtypes = [CInt]
LIB.Chunk_Equals.argtypes = [CInt, CInt]
LIB.Chunk_HighestFilledSubChunk.argtypes = [CInt]
LIB.Chunk_Range.argtypes = [CInt]
LIB.Chunk_SetBiome.argtypes = [CInt, CInt, CInt, CInt, CInt]
LIB.Chunk_SetBlock.argtypes = [CInt, CInt, CInt, CInt, CInt, CInt]
LIB.Chunk_SetBlocks.argtypes = [CInt, CInt, CSlice]
LIB.Chunk_Sub.argtypes = [CInt]
LIB.Chunk_SubChunk.argtypes = [CInt, CInt]
LIB.Chunk_SubIndex.argtypes = [CInt, CInt]
LIB.Chunk_SubY.argtypes = [CInt, CInt]

LIB.NewChunk.restype = CInt
LIB.ReleaseChunk.restype = None
LIB.Chunk_Biome.restype = CInt
LIB.Chunk_Block.restype = CInt
LIB.Chunk_Blocks.restype = CSlice
LIB.Chunk_Compact.restype = CString
LIB.Chunk_Equals.restype = CInt
LIB.Chunk_HighestFilledSubChunk.restype = CInt
LIB.Chunk_Range.restype = CSlice
LIB.Chunk_SetBiome.restype = CString
LIB.Chunk_SetBlock.restype = CString
LIB.Chunk_SetBlocks.restype = CString
LIB.Chunk_Sub.restype = CSlice
LIB.Chunk_SubChunk.restype = CInt
LIB.Chunk_SubIndex.restype = CInt
LIB.Chunk_SubY.restype = CInt


def new_chunk(range_start: int, range_end: int) -> int:
    return int(LIB.NewChunk(CInt(range_start), CInt(range_end)))


def release_chunk(id: int) -> None:
    LIB.ReleaseChunk(CInt(id))


def chunk_biome(id: int, x: int, y: int, z: int) -> int:
    return int(LIB.Chunk_Biome(CInt(id), CInt(x), CInt(y), CInt(z)))


def chunk_block(id: int, x: int, y: int, z: int, layer: int) -> int:
    return int(LIB.Chunk_Block(CInt(id), CInt(x), CInt(y), CInt(z), CInt(layer)))


def chunk_blocks(id: int, layer: int) -> numpy.ndarray:
    return numpy.frombuffer(
        as_python_bytes(LIB.Chunk_Blocks(CInt(id), CInt(layer))), dtype=numpy.uint32
    ).copy()


def chunk_compact(id: int) -> str:
    return as_python_string(LIB.Chunk_Compact(CInt(id)))


def chunk_equals(id: int, another_chunk_id: int) -> int:
    return int(LIB.Chunk_Equals(CInt(id), CInt(another_chunk_id)))


def chunk_highest_filled_sub_chunk(id: int) -> int:
    return int(LIB.Chunk_HighestFilledSubChunk(CInt(id)))


def chunk_range(id: int) -> tuple[int, int, bool]:
    data = as_python_bytes(LIB.Chunk_Range(CInt(id)))
    if len(data) == 0:
        return (0, 0, False)

    start_range = struct.unpack("<i", data[0:4])[0]
    end_range = struct.unpack("<i", data[4:])[0]

    return (start_range, end_range, True)


def chunk_set_biome(id: int, x: int, y: int, z: int, biome_id: int) -> str:
    return as_python_string(
        LIB.Chunk_SetBiome(CInt(id), CInt(x), CInt(y), CInt(z), CInt(biome_id))
    )


def chunk_set_block(
    id: int, x: int, y: int, z: int, layer: int, block_runtime_id: int
) -> str:
    return as_python_string(
        LIB.Chunk_SetBlock(
            CInt(id), CInt(x), CInt(y), CInt(z), CInt(layer), CInt(block_runtime_id)
        )
    )


def chunk_set_blocks(id: int, layer: int, blocks: numpy.ndarray) -> str:
    return as_python_string(
        LIB.Chunk_SetBlocks(CInt(id), CInt(layer), as_c_bytes(blocks.tobytes()))
    )


def chunk_sub(id: int) -> list[int]:
    raw = as_python_bytes(LIB.Chunk_Sub(CInt(id)))
    result = []

    ptr = 0
    while ptr < len(raw):
        result.append(struct.unpack("<I", raw[ptr : ptr + 4])[0])
        ptr += 4

    return result


def chunk_sub_chunk(id: int, y: int) -> int:
    return int(LIB.Chunk_SubChunk(CInt(id), CInt(y)))


def chunk_sub_index(id: int, y: int) -> int:
    return int(LIB.Chunk_SubIndex(CInt(id), CInt(y)))


def chunk_sub_y(id: int, index: int) -> int:
    return int(LIB.Chunk_SubY(CInt(id), CInt(index)))



from io import BytesIO
from struct import unpack
from .getByte import getByte
import nbtlib

endian: str = '<'


class GetValueError(Exception):
    def __init__(self, errorOccurredPosition: int):
        Exception.__init__(
            self, f'failed to parse value, and the error occurred at position {errorOccurredPosition}')


class UnexpectedError(Exception):
    def __init__(self, value: int):
        Exception.__init__(self, f'unexpected number {value}')


def getName(buffer: BytesIO) -> str:
    keyLength = unpack(f'{endian}H', getByte(buffer, 2))[0]
    return getByte(buffer, keyLength).decode(encoding='utf-8', errors="replace")


def getValue(buffer: BytesIO, valueType: int) -> nbtlib.tag.Byte | nbtlib.tag.Short | nbtlib.tag.Int | nbtlib.tag.Long | nbtlib.tag.Float | nbtlib.tag.Double | nbtlib.tag.ByteArray | nbtlib.tag.String | nbtlib.tag.List | nbtlib.tag.Compound | nbtlib.tag.IntArray | nbtlib.tag.LongArray:
    match valueType:
        case 1:
            return nbtlib.tag.Byte(unpack(f'{endian}b', getByte(buffer, 1))[0])
        case 2:
            return nbtlib.tag.Short(unpack(f'{endian}h', getByte(buffer, 2))[0])
        case 3:
            return nbtlib.tag.Int(unpack(f'{endian}i', getByte(buffer, 4))[0])
        case 4:
            return nbtlib.tag.Long(unpack(f'{endian}q', getByte(buffer, 8))[0])
        case 5:
            return nbtlib.tag.Float(unpack(f'{endian}f', getByte(buffer, 4))[0])
        case 6:
            return nbtlib.tag.Double(unpack(f'{endian}d', getByte(buffer, 8))[0])
        case 7 | 11 | 12:
            return getArray(buffer, valueType)
        case 8:
            return nbtlib.tag.String(getName(buffer))
        case 9:
            return getList(buffer)
        case 10:
            return getCompound(buffer)
        case _:
            raise GetValueError(buffer.seek(0, 1))


def getArray(buffer: BytesIO, valueType: int) -> nbtlib.tag.ByteArray | nbtlib.tag.IntArray | nbtlib.tag.LongArray:
    arrayLength = unpack(f'{endian}i', getByte(buffer, 4))[0]
    match valueType:
        case 7:
            return nbtlib.tag.ByteArray([unpack(f'{endian}b', getByte(buffer, 1))[0] for _ in range(arrayLength)])
        case 11:
            return nbtlib.tag.IntArray([unpack(f'{endian}i', getByte(buffer, 4))[0] for _ in range(arrayLength)])
        case 12:
            return nbtlib.tag.LongArray([unpack(f'{endian}q', getByte(buffer, 8))[0] for _ in range(arrayLength)])
        case _:
            raise UnexpectedError(valueType)


def getList(buffer: BytesIO) -> nbtlib.tag.List:
    valueType = getByte(buffer, 1)[0]
    listLength = unpack(f'{endian}i', getByte(buffer, 4))[0]
    return nbtlib.tag.List([getValue(buffer, valueType) for _ in range(listLength)])


def getCompound(buffer: BytesIO) -> nbtlib.tag.Compound:
    result: dict = {}
    while True:
        nextBuffer = getByte(buffer, 1)
        if nextBuffer == b'\x00':
            return nbtlib.tag.Compound(result)
        # if meet TAG_End
        buffer.seek(-1, 1)
        # correct the pointer
        valueType = getByte(buffer, 1)[0]
        name = getName(buffer)
        result[name] = getValue(buffer, valueType)
        # set values


def UnMarshalBufferToPythonNBTObject(buffer: BytesIO) -> tuple[nbtlib.tag.Byte | nbtlib.tag.Short | nbtlib.tag.Int | nbtlib.tag.Long | nbtlib.tag.Float | nbtlib.tag.Double | nbtlib.tag.ByteArray | nbtlib.tag.String | nbtlib.tag.List | nbtlib.tag.Compound | nbtlib.tag.IntArray | nbtlib.tag.LongArray, str]:
    valueType = getByte(buffer, 1)[0]
    name = getName(buffer)
    return getValue(buffer, valueType), name



import nbtlib
from .constant import (
    AIR_BLOCK_STATES,
    EMPTY_BLOCK_STATES,
    RANGE_OVERWORLD,
)
from .define import BlockStates, Range
from ..world.sub_chunk import SubChunk, SubChunkWithIndex
from ..internal.symbol_export_conversion import (
    runtime_id_to_state as rits,
    state_to_runtime_id as stri,
    sub_chunk_network_payload as scnp,
    from_sub_chunk_network_payload as fscnp,
    sub_chunk_disk_payload as scdp,
    from_sub_chunk_disk_payload as fscdp,
)


def runtime_id_to_state(
    block_runtime_id: int,
) -> BlockStates:
    """runtime_id_to_state convert block runtime id to a BlockStates.

    Args:
        block_runtime_id (int): The runtime id of target block.

    Returns:
        BlockStates: If not found, return AIR_BLOCK_STATES.
                     Otherwise, return the founded block states.
    """
    block_states = BlockStates()

    name, states, success = rits(block_runtime_id)
    if not success:
        return AIR_BLOCK_STATES

    block_states.Name, block_states.States = name, states  # type: ignore
    return block_states


def state_to_runtime_id(
    block_name: str, block_states: nbtlib.tag.Compound = EMPTY_BLOCK_STATES
) -> int:
    """
    state_to_runtime_id convert a block which name is block_name
    and states is block_states to its block runtime id represent.

    Args:
        block_name (str): The name of this block.
        block_states (nbtlib.tag.Compound, optional): The block states of this block.
                                                      Defaults to EMPTY_BLOCK_STATES.

    Returns:
        int: If not found, return 0.
             Otherwise, return its block runtime id.
    """
    block_runtime_id, success = stri(block_name, block_states)
    if not success:
        return 0
    return block_runtime_id


def sub_chunk_network_payload(
    sub_chunk: SubChunk, index: int, r: Range = RANGE_OVERWORLD
) -> bytes:
    """
    sub_chunk_network_payload encodes sub_chunk to its payload represent that could use on network sending.

    Args:
        sub_chunk (SubChunk): The sub chunk want to encode.
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0
        r (Range, optional): The whole chunk range where this sub chunk is in.
                             For overworld, it is Range(-64, 319).
                             Defaults to RANGE_OVERWORLD.


    Returns:
        bytes: The bytes represent of this sub chunk, and could especially sending on network.
               Therefore, this is a Network encoding sub chunk payload.
    """
    return scnp(sub_chunk._sub_chunk_id, r.start_range, r.end_range, index)


def from_sub_chunk_network_payload(
    payload: bytes, r: Range = RANGE_OVERWORLD
) -> SubChunkWithIndex:
    """from_sub_chunk_network_payload decode a Network encoding sub chunk and return its python represent.

    Args:
        payload (bytes): The bytes of this sub chunk, who with a Network encoding.
        r (Range, optional): The whole chunk range where this sub chunk is in.
                             For overworld, it is Range(-64, 319).
                             Defaults to RANGE_OVERWORLD.

    Returns:
        SubChunkWithIndex: If failed to decode, then return a invalid sub chunk and a invalid -1 sub chunk Y index.
                           Otherwise, return decoded sub chunk with its Y index.
                           Note that you could use s.sub_chunk.is_valid() to check whether the sub chunk is valid or not.
    """
    s = SubChunkWithIndex(-1)
    index, sub_chunk_id, success = fscnp(r.start_range, r.end_range, payload)
    if not success:
        return s
    s.index, s.sub_chunk._sub_chunk_id = index, sub_chunk_id
    return s


def sub_chunk_disk_payload(
    sub_chunk: SubChunk, index: int, r: Range = RANGE_OVERWORLD
) -> bytes:
    """
    sub_chunk_disk_payload encodes sub_chunk to its payload represent under Disk encoding.
    That means the returned bytes could save to disk if its len bigger than 0.

    Args:
        sub_chunk (SubChunk): The sub chunk want to encode.
        index (int): The index of this sub chunk, must bigger than -1.
                     For example, for a block in (x, -63, z), than its
                     sub chunk Y pos will be -63>>4 (-4).
                     However, this is not the index of this sub chunk,
                     we need do other compute to get the index:
                     index = (-63>>4) - (r.start_range>>4)
                           = (-63>>4) - (-64>>4)
                           = 0
        r (Range, optional): The whole chunk range where this sub chunk is in.
                             For overworld, it is Range(-64, 319).
                             Defaults to RANGE_OVERWORLD.


    Returns:
        bytes: The bytes represent of this sub chunk, who with a Disk encoding.
    """
    return scdp(sub_chunk._sub_chunk_id, r.start_range, r.end_range, index)


def from_sub_chunk_disk_payload(
    payload: bytes, r: Range = RANGE_OVERWORLD
) -> SubChunkWithIndex:
    """from_sub_chunk_disk_payload decode a Disk encoding sub chunk and return its python represent.

    Args:
        payload (bytes): The bytes of this sub chunk, who with a Disk encoding.
        r (Range, optional): The whole chunk range where this sub chunk is in.
                             For overworld, it is Range(-64, 319).
                             Defaults to RANGE_OVERWORLD.

    Returns:
        SubChunkWithIndex: If failed to decode, then return a invalid sub chunk and a invalid -1 sub chunk Y index.
                           Otherwise, return decoded sub chunk with its Y index.
                           Note that you could use s.sub_chunk.is_valid() to check whether the sub chunk is valid or not.
    """
    s = SubChunkWithIndex(-1)
    index, sub_chunk_id, success = fscdp(r.start_range, r.end_range, payload)
    if not success:
        return s
    s.index, s.sub_chunk._sub_chunk_id = index, sub_chunk_id
    return s



import setuptools

with open("requirements.txt", "r+", encoding="utf-8") as file:
    dependences = file.read().strip().split("\n")

with open("README.md", "r+", encoding="utf-8") as file:
    long_description = file.read()

with open("version", "r+", encoding="utf-8") as file:
    version = file.read().strip()


setuptools.setup(
    name="bedrock-world-operator",
    version=version,
    author="Minecraft Muti-Media Organization",
    author_email="TriM-Organization@hotmail.com",
    description="An operator based on Go that aims to provide interface for Python that could operating NetEase/Standard Minecraft bedrock game saves.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TriM-Organization/bedrock-world-operator",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Go",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
    ],
    package_dir={"bedrockworldoperator": "python/package"},
    packages=[
        i.replace("package", "bedrockworldoperator", 1)
        for i in setuptools.find_packages(where="python")
    ],
    package_data={
        "bedrockworldoperator": [
            "dynamic_libs/*.so",
            "dynamic_libs/*.dll",
            "dynamic_libs/*.dylib",
        ],
    },
    install_requires=dependences,
    python_requires=">=3.11",
)



from io import BytesIO
from .getByte import getByte


def getString(buffer: BytesIO) -> str:
    writer = BytesIO(b'')
    while True:
        currentByte = getByte(buffer, 1)
        if currentByte == b'\x00':
            return writer.getvalue().decode(encoding='utf-8', errors='replace')
        else:
            writer.write(currentByte)



import nbtlib
from ..internal.symbol_export_conversion import state_to_runtime_id
from ..world.define import BlockStates, Dimension, Range

EMPTY_COMPOUND = nbtlib.tag.Compound()
EMPTY_BLOCK_STATES = EMPTY_COMPOUND

DIMENSION_ID_OVERWORLD = 0
DIMENSION_ID_NETHER = 1
DIMENSION_ID_END = 2

DIMENSION_OVERWORLD = Dimension(DIMENSION_ID_OVERWORLD)
DIMENSION_NETHER = Dimension(DIMENSION_ID_NETHER)
DIMENSION_END = Dimension(DIMENSION_ID_END)

RANGE_OVERWORLD = DIMENSION_OVERWORLD.range()
RANGE_NETHER = DIMENSION_NETHER.range()
RANGE_END = DIMENSION_END.range()
RANGE_INVALID = Range(0, -1)

AIR_BLOCK_STATES = BlockStates("minecraft:air")
AIR_BLOCK_RUNTIME_ID = state_to_runtime_id("minecraft:air", EMPTY_BLOCK_STATES)[0]
